/* meko
 * Copyright (C) 2001 Scott James Remnant <scott@netsplit.com>.
 * All Rights Reserved.
 *
 * chat_interface.c
 *  - Code to create basic chat interface which can be packed into
 *    other windows or notebooks
 *  - Maintains list of available chat interfaces linked to servers, users
 *    or channels
 *  - Functions to output to, or update chat interfaces based on recipient
 * -- 
 * $Id: chat_interface.c,v 1.20 2001/06/17 17:53:27 scott Exp $
 * 
 * This file is distributed according to the GNU General Public
 * License.  For full details, read the top of 'main.c' or the
 * file called COPYING that was distributed with this code.
 */

#include <meko.h>

#include <string.h>

#include <glib.h>
#include <gnome.h>
#include <gdk/gdk.h>
#include <gtk/gtk.h>
#include <gdk-pixbuf/gdk-pixbuf.h>

#include "chat_interface.h"
#include "say2_types.h"
#include "say2_server.h"
#include "say2_client.h"
#include "user_commands.h"

/* Forward prototypes for callbacks */
static void     chat_interface_destroy(GtkWidget *, gpointer);
static void     chat_interface_entry_activated(GtkWidget *, ChatInterface *);
static gboolean chat_interface_entry_key_pressed(GtkWidget *, GdkEventKey *,
						 ChatInterface *);

/* Forward prototypes for other functions */
static void chat_interface_sayv(Say2Server *, const gchar *, va_list);
static void chat_interface_saytov(ChatInterface *, const gchar *, va_list);
static void chat_interface_unconfig(ChatInterface *);
static void chat_interface_config(ChatInterface *);
static void chat_interface_update_labels(ChatInterface *);
static void chat_interface_update_entry(ChatInterface *, gint, Say2Client *);

/* Font for the interface */
#define CHAT_FONT ("-*-lucidatypewriter-medium-r" \
		   "-normal-*-*-120-*-*-*-*-iso8859-1")

/* List of currently open interfaces */
static GSList *open_interfaces = NULL;


/* chat_interface_entry_activated
 * Callback for when the user activates the entry box (by pressing Enter in
 * it).  Wraps around the command handler.
 */
static void
chat_interface_entry_activated(GtkWidget *widget, ChatInterface *ci)
{
	const gchar *text;

	g_return_if_fail(widget);
	g_return_if_fail(ci);

	/* Get the text they typed into the widget */
	text = gtk_entry_get_text(GTK_ENTRY(widget));

	/* Handle the text */
	user_handle_text(ci, text);

	/* Clear the widget */
	gtk_entry_set_text(GTK_ENTRY(widget), "");
}

/* chat_interface_entry_key_pressed
 * Callback for when the user pushes a key in the entry box.  Look for up and
 * down arrows, and tab too.  Then do either history stuff or tab completion
 * stuff.
 */
static gboolean
chat_interface_entry_key_pressed(GtkWidget *widget, GdkEventKey *event,
				 ChatInterface *ci)
{
	g_return_val_if_fail(widget, FALSE);
	g_return_val_if_fail(event, FALSE);
	g_return_val_if_fail(ci, FALSE);

	/* Switch on key pressed */
	switch(event->keyval) {
	case GDK_Tab:
		/* Tab key.  Do tab completion stuff. */
		g_print("TAB!\n");
		break;

	case GDK_Up:
		/* Up key.  Do history recall stuff */
		g_print("UP!\n");
		break;

	case GDK_Down:
		/* Down key.  Do history recall stuff */
		g_print("DOWN!\n");
		break;

	case GDK_Page_Up:
		/* Page Up key.  Scroll up the output box */
		g_print("PAGEUP!\n");
		break;

	case GDK_Page_Down:
		/* Page Down key.  Scroll down the output box */
		g_print("PAGEDOWN!\n");
		break;

	default:
		/* Unhandled key */
		return FALSE;
	}

	/* If we get here then we handled the key, stop the signal being
	 * used to do other things
	 */
	gtk_signal_emit_stop_by_name(GTK_OBJECT(widget), "key-press-event");
	return TRUE;
}

/* chat_interface_saytov
 * Output formatted information to a specific chat interface
 */
static void
chat_interface_saytov(ChatInterface *ci, const gchar *msg, va_list args)
{
	gchar *text;

	g_return_if_fail(ci);

	text = g_strdup_vprintf(msg, args);
	gtk_text_insert(GTK_TEXT(ci->output_box), NULL, NULL, NULL,
			text, strlen(text));
	g_free(text);
}

/* chat_interface_sayv
 * Output formatted information to all chat interfaces bound to the server
 * given.  FIXME this should really include a RECIPTYPE and RECIP to indicate
 * exactly which interfaces we should output it to.
 */
static void
chat_interface_sayv(Say2Server *server, const gchar *msg, va_list args)
{
	GSList *list;

	/* Iterate away */
	for (list = open_interfaces; list; list = g_slist_next(list)) {
		ChatInterface *ci;

		ci = (ChatInterface *)list->data;
		g_assert(ci);

		/* Skip any that don't match */
		if (ci->server != server)
			continue;

		/* Just call saytov */
		chat_interface_saytov(ci, msg, args);
	}
}

/* chat_interface_sayto
 * Output formatted information to a specific chat interface
 */
void
chat_interface_sayto(ChatInterface *ci, const gchar *msg, ...)
{
	va_list args;

	/* Just a wrapper around saytov */
	va_start(args, msg);
	chat_interface_saytov(ci, msg, args);
	va_end(args);
}

/* chat_interface_say
 * Output formatted information to all chat interfaces bound to the server
 * given.  FIXME like we fix sayv
 */
void
chat_interface_say(Say2Server *server, const gchar *msg, ...)
{
	va_list args;

	/* Just a wrapper around sayv */
	va_start(args, msg);
	chat_interface_sayv(server, msg, args);
	va_end(args);
}

/* chat_interface_disconnect
 * Disconnect the specified chat interface from whatever it is bound to and
 * return it to the unused pool.
 */
void
chat_interface_disconnect(ChatInterface *ci)
{
	if (!ci)
		return;

	/* Unreference the server */
	say2_server_unref(ci->server);
	ci->server = NULL;

	/* Unconfigure the interface */
	chat_interface_unconfig(ci);

	/* Unreference and reset info about the recipient */
	say2_client_unref(ci->recip);
	ci->recip = 0;
	ci->recip_type = CHAT_RECIPTYPE_ANY;

	/* Reconfigure the interface */
	chat_interface_config(ci);
}

/* chat_interface_update
 * This is the function that most stuff calls to update chat interfaces with
 * new information from the say2 server.  It acts on all ChatInterface's with
 * the Say2Server specified, and takes the type of update and a Say2Client
 * containing the new information for the client being updated.
 */
void
chat_interface_update(Say2Server *server, ChatUpdateType update_type,
		      Say2Client *client)
{
	GSList      *list;
	gint         row;

	/* Interate the lists */
	for (list = open_interfaces; list; list = g_slist_next(list)) {
		ChatInterface *ci;

		ci = (ChatInterface *)list->data;
		g_assert(ci);

		/* Skip any that don't match */
		if (ci->server != server)
			continue;

		switch (update_type) {
		case CHAT_UPDATE_CLOSE:
			/* Indicates that the server the Interface is
			 * connected to is about to be closed */
			chat_interface_disconnect(ci);
			break;

		case CHAT_UPDATE_SIGNON:
			/* Indicates that the user has just signed on to the
			 * channel (or needs adding to it, anyway)
			 */
			g_return_if_fail(client);
			row = gtk_clist_find_row_from_data
				(GTK_CLIST(ci->user_list), client);
			if (row == -1) {
				gchar *blanks[] = { "", "", "" };

				/* Append a blank row, setting the height and
				 * data correctly
				 */
				row = gtk_clist_append
					(GTK_CLIST(ci->user_list), blanks);
				gtk_clist_set_row_height
					(GTK_CLIST(ci->user_list), 32);
				gtk_clist_set_row_data
					(GTK_CLIST(ci->user_list), row, client);
			}
		
			/*
			 * FALLTHROUGH!
			 * Everything else can be done by USER
			 */
		case CHAT_UPDATE_USER:
			/* Indicates that information about a user needs
			 * updating (including us, so it may not be in the
			 * channel) */
			g_return_if_fail(client);
			row = gtk_clist_find_row_from_data
				(GTK_CLIST(ci->user_list), client);
			if (row == -1)
				break;

			/* Update the entry */
			chat_interface_update_entry(ci, row, client);

			/* Sort it */
			gtk_clist_sort(GTK_CLIST(ci->user_list));

			break;
			
		case CHAT_UPDATE_SIGNOFF:
			/* Indicates that a user has left the channel and 
			 * needs removing from the user list
			 */
			g_return_if_fail(client);
			row = gtk_clist_find_row_from_data
				(GTK_CLIST(ci->user_list), client);
			if (row == -1)
				break;

			/* Remove the row */
			gtk_clist_remove(GTK_CLIST(ci->user_list), row);

			break;

		default:
			/* NOTREACHED! */
			g_assert_not_reached();
		}

		/* It's a good bet that something's been changed that'll
		 * warrant us doing this
		 */
		chat_interface_update_labels(ci);
	}
}

/* chat_interface_update_entry
 * Update a list entry based on the information passed
 */
static void
chat_interface_update_entry(ChatInterface *ci, gint row, Say2Client *client)
{
	static gint  imgnum = 0;
	const gchar *nick;
	gchar       *sort_key;

	g_return_if_fail(ci);
	g_return_if_fail(client);

	/* Work out what the nickname and sort key will be */
	nick = (client->nick ? client->nick : client->id);
	g_strdown(sort_key = g_strdup(nick));

	/* Fill in the text */
	gtk_clist_set_text(GTK_CLIST(ci->user_list), row, 1, nick);
	gtk_clist_set_text(GTK_CLIST(ci->user_list), row, 2, sort_key);

	/* Pick a random icon if they don't have one for now.  If they do have
	 * one we actually ignore it at the moment.  (we set icon to -1 to
	 * prevent their icon changing every minute)
	 */
	if (!client->icon) {
		GdkPixbuf *pixbuf;
		gchar     *filename;

		/* This should point to the images directory in the mochi
		 * source for now.  Look, it's a hack ok
		 */
		filename = g_strdup_printf("images/person%d.gif", imgnum + 1);
		imgnum++;
		if (imgnum == 16)
			imgnum = 0;

		/* Use gdk-pixbuf to load the file, this'll load most common
		 * formats correctly
		 */
		pixbuf = gdk_pixbuf_new_from_file(filename);
		if (pixbuf) {
			GdkPixmap *pixmap;
			GdkBitmap *mask;

			/* Render the pixbuf onto a normal Pixmap and Mask */
			gdk_pixbuf_render_pixmap_and_mask(pixbuf, &pixmap,
							  &mask, 0);

			/* Set the pixmap of the list entry */
			gtk_clist_set_pixmap(GTK_CLIST(ci->user_list), row, 0,
					    pixmap, mask);

			/* Unreference the pixbuf, so it'll be freed */
			gdk_pixbuf_unref(pixbuf);
		}

		/* Free the filename */
		g_free(filename);

		/* Don't do this for this entry anymore */
		client->icon = (char *)-1;
	}

	/* Free up */
	g_free(sort_key);
}

/* chat_interface_get
 * Search the open_interfaces list for a chat interface matching the
 * parameters specified, or one that isn't being used that we could make
 * match the parameters specified.  If all else fails, return NULL so the
 * called knows to create one.
 */
ChatInterface *
chat_interface_get(Say2Server *server, ChatRecipType recip_type,
		   Say2Client *recip)
{
	ChatInterface *likely[3], *ci;
	GSList        *list;
	gint           i;

	/* Scan the open interfaces list only once, finding candidates.
	 * likely[0] will be an exact, active match.
	 * likely[1] will be an inactive interface on the current server
	 * likely[2] will be a completley inactive interface
	 */
	likely[0] = likely[1] = likely[2] = NULL;
	for (list = open_interfaces; list; list = g_slist_next(list)) {
		ci = (ChatInterface *)list->data;
		g_assert(ci);

		if (!likely[0] && (ci->server == server)
		    && (ci->recip_type == recip_type) && (ci->recip == recip))
			likely[0] = ci;

		if (!likely[1] && (ci->server == server)
		    && (ci->recip_type == CHAT_RECIPTYPE_ANY))
			likely[1] = ci;

		if (!likely[2] && !ci->server)
			likely[2] = ci;
	}

	/* Pick the best one */
	for (i = 0; i < 3; i++) {
		ci = likely[i];
		if (ci)
			break;
	}

	/* Return NULL to indicate that the caller should create one */
	if (!ci)
		return NULL;

	/* Set the chat interface up and return it */
	chat_interface_set(ci, server, recip_type, recip);
	return ci;
}

/* chat_interface_set
 * Set the specified chat interface up with the information given.
 */
void
chat_interface_set(ChatInterface *ci, Say2Server *server,
		   ChatRecipType recip_type, Say2Client *recip)
{
	g_return_if_fail(ci);

	/* Unconfigure the window */
	chat_interface_unconfig(ci);

	/* Link the server, unless it's already the same one */
	if (ci->server != server) {
		ci->server = server;
		say2_server_ref(ci->server);
	}

	/* Link the recipient, unless it's already the same */
	if (ci->recip != recip) {
		ci->recip = recip;
		say2_client_ref(ci->recip);
	}
	ci->recip_type = recip_type;

	/* Reconfigure the window */
	chat_interface_config(ci);

	/* Update the usual suspects (nick, title, etc) */
	chat_interface_update_labels(ci);
}

/* chat_interface_unconfig
 * Make sure that all the movable widgets are unparented and hidden.  This
 * MUST be followed by a chat_interface_config before almost anything else
 * is done.
 */
static void
chat_interface_unconfig(ChatInterface *ci)
{
	g_return_if_fail(ci);

	/* Hide the widgets */
	gtk_widget_hide(ci->whois_FIXME);
	gtk_widget_hide(ci->top_pane);
	gtk_widget_hide(ci->output_hbox);

	/* Reference them all so they don't get magically destroyed by GTK+ */
	gtk_widget_ref(ci->whois_FIXME);
	gtk_widget_ref(ci->top_pane);
	gtk_widget_ref(ci->output_hbox);
		
	/* De-parent them, we know where most of them are, but will need to
	 * have a good guess about the output_hbox
	 */
	gtk_container_remove(GTK_CONTAINER(ci->top_vbox), ci->whois_FIXME);
	gtk_container_remove(GTK_CONTAINER(ci->top_vbox), ci->top_pane);
	if (ci->recip_type == CHAT_RECIPTYPE_CHAN) {
		gtk_container_remove(GTK_CONTAINER(ci->top_pane),
				     ci->output_hbox);
	} else {
		gtk_container_remove(GTK_CONTAINER(ci->top_vbox),
				     ci->output_hbox);
	}
}

/* chat_interface_config
 * Reconfigure an interface previously deconfigured by chat_interface_deconfig,
 * reparenting all the widgets in the right place and showing the ones we need
 * for this interface
 */
static void
chat_interface_config(ChatInterface *ci)
{
	g_return_if_fail(ci);

	/* Reparent all the widgets, output_hbox will need to go in one of
	 * two places
	 */
	gtk_box_pack_start(GTK_BOX(ci->top_vbox), ci->whois_FIXME,
			   FALSE, FALSE, 0);
	if (ci->recip_type == CHAT_RECIPTYPE_CHAN) {
		gtk_paned_pack1(GTK_PANED(ci->top_pane), ci->output_hbox,
				TRUE, TRUE);
	} else {
		gtk_box_pack_start(GTK_BOX(ci->top_vbox), ci->output_hbox,
				   TRUE, TRUE, 0);
	}
	gtk_box_pack_start(GTK_BOX(ci->top_vbox), ci->top_pane, TRUE, TRUE, 0);

	/* Unreference the widgets, they're all referenced by their parents
	 * so they won't be freed now
	 */
	gtk_widget_unref(ci->whois_FIXME);
	gtk_widget_unref(ci->top_pane);
	gtk_widget_unref(ci->output_hbox);

	/* Any other business */
	gtk_clist_clear(GTK_CLIST(ci->user_list));

	/* Show the right widgets */
	if (ci->recip_type == CHAT_RECIPTYPE_USER)
		gtk_widget_show(ci->whois_FIXME);
	gtk_widget_show(ci->output_hbox);
	if (ci->recip_type == CHAT_RECIPTYPE_CHAN)
		gtk_widget_show(ci->top_pane);
}

/* chat_interface_update_labels
 * Update the various labels on the interface and call the title callback
 */
static void
chat_interface_update_labels(ChatInterface *ci)
{
	g_return_if_fail(ci);

	/* Update the nickname label.  Hide and show it as necessary */
	if (ci->server) {
		const gchar *text;

		text = (ci->server->me->nick ? ci->server->me->nick
					     : ci->server->me->id);
		gtk_label_set_text(GTK_LABEL(ci->nick_label), text);
		gtk_widget_show(ci->nick_label);
	} else {
		gtk_widget_hide(ci->nick_label);
	}

	/* Call the title callback */
	if (ci->title_cb)
		ci->title_cb(ci, ci->title_cb_data);

	/* Always a good bet to do this */
	gtk_widget_grab_focus(ci->entry_box);
}

/* chat_interface_set_title_cb
 * Set a function which will be called when the chat interface thinks that
 * you might want to change the title of its parent window or notebook etc
 */
void
chat_interface_set_title_cb(ChatInterface *ci, ChatTitleFn title_cb,
			    gpointer title_cb_data)
{
	g_return_if_fail(ci);

	ci->title_cb = title_cb;
	ci->title_cb_data = title_cb_data;
}

/* chat_interface_destroy
 * Called when the top vbox is destroyed, use to clean up this interface
 */
static void
chat_interface_destroy(GtkWidget *widget, gpointer data)
{
	ChatInterface *ci;

	ci = gtk_object_get_data(GTK_OBJECT(widget), "chat_interface");
	g_return_if_fail(ci);

	/* Not using these anymore */
	say2_server_unref(ci->server);
	say2_client_unref(ci->recip);

	/* Remove ourselves from the list of open interfaces */
	open_interfaces = g_slist_remove(open_interfaces, ci);

	/* Free the interface */
	g_free(ci);
}

/* chat_interface_new
 * Create a new chat interface, returns the structure of information for
 * storage or futher manipulation
 */
ChatInterface *
chat_interface_new(void)
{
	ChatInterface *ci;
	GtkWidget     *hbox, *scrollwin, *scroll;
	GdkFont       *chat_font;
	GtkStyle      *style;

	/* Load the font */
	chat_font = gdk_font_load(CHAT_FONT);

	/* Create the interface and add it to the list */
	ci = g_new0(ChatInterface, 1);
	ci->recip_type = CHAT_RECIPTYPE_ANY;
	open_interfaces = g_slist_prepend(open_interfaces, ci);

	/* Create a box for the elements.  We hook the destroy signal for this
	 * so that we can free everything up safely without worrying about
	 * our parent remembering
	 */
	ci->top_vbox = gtk_vbox_new(FALSE, 2);
	gtk_object_set_data(GTK_OBJECT(ci->top_vbox), "chat_interface", ci);
	gtk_signal_connect(GTK_OBJECT(ci->top_vbox), "destroy",
			   GTK_SIGNAL_FUNC(chat_interface_destroy), NULL);
	gtk_widget_show(ci->top_vbox);

	
	/* Whois information.  FIXME */
	ci->whois_FIXME = gtk_label_new("Whois information goes here");
	gtk_box_pack_start(GTK_BOX(ci->top_vbox), ci->whois_FIXME,
			   TRUE, TRUE, 0);


	/* Paned window to hold output box and user list.  We create this
	 * invisible at first, and repack it all later when we're told to be
	 * a CHAN interface
	 */
	ci->top_pane = gtk_hpaned_new();
	gtk_box_pack_start(GTK_BOX(ci->top_vbox), ci->top_pane, TRUE, TRUE, 0);

	/* Create the user list scrolling window on the right */
	scrollwin = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrollwin),
				       GTK_POLICY_AUTOMATIC,
				       GTK_POLICY_AUTOMATIC);
	gtk_paned_pack2(GTK_PANED(ci->top_pane), scrollwin, FALSE, TRUE);
	gtk_widget_show(scrollwin);

	/* Create a user list */
	ci->user_list = gtk_clist_new(3);
	gtk_clist_set_column_width(GTK_CLIST(ci->user_list), 0, 32);
	gtk_clist_set_column_auto_resize(GTK_CLIST(ci->user_list), 1, TRUE);
	gtk_clist_set_column_visibility(GTK_CLIST(ci->user_list), 2, FALSE);
	gtk_clist_set_selection_mode(GTK_CLIST(ci->user_list),
			             GTK_SELECTION_MULTIPLE);
	gtk_clist_set_sort_column(GTK_CLIST(ci->user_list), 2);
	gtk_clist_set_sort_type(GTK_CLIST(ci->user_list), GTK_SORT_ASCENDING);
	gtk_widget_set_usize(ci->user_list, 140, -1);
	gtk_container_add(GTK_CONTAINER(scrollwin), ci->user_list);
	gtk_widget_show(ci->user_list);


	/* Create a box for the output widget and its scrollbar.  For now this
         * goes in the top, but we might move it into the pane later
	 */
	ci->output_hbox = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(ci->top_vbox), ci->output_hbox,
			   TRUE, TRUE, 0);
	gtk_widget_show(ci->output_hbox);

	/* Create a text box for the output */
	ci->output_box = gtk_text_new(NULL, NULL);
	gtk_text_set_editable(GTK_TEXT(ci->output_box), FALSE);
	gtk_text_set_word_wrap(GTK_TEXT(ci->output_box), TRUE);
	gtk_box_pack_start(GTK_BOX(ci->output_hbox), ci->output_box,
			   TRUE, TRUE, 0);
	gtk_widget_show(ci->output_box);

	/* Restyle the text box */
	style = gtk_style_copy(gtk_widget_get_style(ci->output_box));
	style->font = chat_font;
	gtk_widget_set_style(ci->output_box, style);

	/* Create scroll bar for the text box */
	scroll = gtk_vscrollbar_new(GTK_TEXT(ci->output_box)->vadj);
	gtk_box_pack_end(GTK_BOX(ci->output_hbox), scroll, FALSE, FALSE, 0);
	gtk_widget_show(scroll);

	/* Create a box for the nick label and the entry box */
	hbox = gtk_hbox_new(FALSE, 2);
	gtk_box_pack_end(GTK_BOX(ci->top_vbox), hbox, FALSE, FALSE, 0);
	gtk_widget_show(hbox);

	/* Create a label for the nickname */
	ci->nick_label = gtk_label_new("");
	gtk_box_pack_start(GTK_BOX(hbox), ci->nick_label, FALSE, FALSE, 0);

	/* Create an entry box */
	ci->entry_box = gtk_entry_new();
	gtk_signal_connect(GTK_OBJECT(ci->entry_box), "activate",
			   GTK_SIGNAL_FUNC(chat_interface_entry_activated),
			   (gpointer)ci);
	gtk_signal_connect(GTK_OBJECT(ci->entry_box), "key-press-event",
			   GTK_SIGNAL_FUNC(chat_interface_entry_key_pressed),
			   (gpointer)ci);
	gtk_box_pack_end(GTK_BOX(hbox), ci->entry_box, TRUE, TRUE, 0);
	gtk_widget_show(ci->entry_box);

	/* Restyle the entry box */
	style = gtk_style_copy(gtk_widget_get_style(ci->entry_box));
	style->font = chat_font;
	gtk_widget_set_style(ci->entry_box, style);

	return ci;
}
