/* meko
 * Copyright (C) 2001 Scott James Remnant <scott@netsplit.com>.
 * All Rights Reserved.
 *
 * connect_dialog.c
 *  - Dialog to ask user for connection information
 *  - Holds connection information in a list of them which are saved between
 *    each session
 * -- 
 * $Id: connect_dialog.c,v 1.13 2001/06/17 18:22:13 scott Exp $
 * 
 * This file is distributed according to the GNU General Public
 * License.  For full details, read the top of 'main.c' or the
 * file called COPYING that was distributed with this code.
 */

#include <meko.h>

#include <glib.h>
#include <gnome.h>
#include <gtk/gtk.h>

#include "connect_dialog.h"
#include "say2_server.h"
#include "chat_interface.h"

/* Structure holding information about a saved connection */
typedef struct _connect_info ConnectInfo;
struct _connect_info {
	/* Name of this connection */
	gchar *name;

	/* Server name and port */
	gchar *server_name;
	gchar *server_port;

	/* Nickname (or id) to use */
	gchar *nick;

	/* Information needed to login */
	gint   as_guest;
	gchar *server_pass;
	gchar *email;
	gchar *tagline;
};

/* Structure holding the important bits of the dialog */
typedef struct _connect_dialog ConnectDialog;
struct _connect_dialog {
	/* The dialog itself */
	GtkWidget *dialog;

	/* Important widgets */
	GtkWidget *frame_vbox;
	GtkWidget *clist;
	GtkWidget *remove_button;
	GtkWidget *conn_name;
	GtkWidget *server_name;
	GtkWidget *server_port;
	GtkWidget *nick;
	GtkWidget *as_user;
	GtkWidget *server_pass;
	GtkWidget *server_pass_label;
	GtkWidget *as_guest;
	GtkWidget *email;
	GtkWidget *email_label;
	GtkWidget *tagline;
	GtkWidget *tagline_label;

	/* List of saved connections */
	GSList *saved;

	/* Currently selected info */
	ConnectInfo *current;
};

/* Forward Prototypes */
static void connect_dialog_load(ConnectDialog *);
static void connect_dialog_save(ConnectDialog *);
static void connect_dialog_apply_clicked(GtkWidget *, ConnectDialog *);
static void connect_dialog_ok_clicked(GtkWidget *, ConnectDialog *);
static void connect_dialog_new_clicked(GtkWidget *, ConnectDialog *);
static void connect_dialog_select(ConnectDialog *, ConnectInfo *);
static void connect_dialog_row_selected(GtkWidget *, gint, gint,
					GdkEventButton *, ConnectDialog *);
static void connect_dialog_row_unselected(GtkWidget *, gint, gint,
					  GdkEventButton *, ConnectDialog *);
static void connect_dialog_remove_clicked(GtkWidget *, ConnectDialog *);
static gint connect_dialog_button_pressed(GtkWidget *, GdkEventButton *,
					  ConnectDialog *);
static void connect_dialog_entry_changed(GtkWidget *, ConnectDialog *);
static void connect_dialog_login_toggled(GtkWidget *, ConnectDialog *);
static void connect_dialog_destroy(GtkWidget *, ConnectDialog **);
static void connect_dialog_free_info(ConnectInfo *);


/* connect_dialog_load
 * Load saved settings out of the configuration database and populate the list
 * with them.
 */
static void
connect_dialog_load(ConnectDialog *cd)
{
	gint num_saved, s;

	g_return_if_fail(cd);

	/* Everything's under here */
	gnome_config_push_prefix("/" APPNAME "/saved-connections/");

	/* Number of saved connections */
	num_saved = gnome_config_private_get_int("number=0");

	/* Iterate the connections */
	for (s = 0; s < num_saved; s++) {
		ConnectInfo *cinfo;
		gchar       *secname;
		gchar       *names[1];
		gint         row;
	
		/* Connections get their own section */
		secname = g_strdup_printf("/%s/saved-connection-%d/",
					  APPNAME, s);
		gnome_config_push_prefix(secname);
		
		/* Allocate space and copy the values into it */
		cinfo = g_new0(ConnectInfo, 1);
		cinfo->name = gnome_config_private_get_string("name=");
		cinfo->server_name = gnome_config_private_get_string
			("server-name=");
		cinfo->server_port = gnome_config_private_get_string
			("server-port=");
		cinfo->nick = gnome_config_private_get_string("nick=");
		cinfo->as_guest = gnome_config_private_get_bool
			("as_guest=FALSE");
		cinfo->server_pass = gnome_config_private_get_string
			("server-pass=");
		cinfo->email = gnome_config_private_get_string("email=");
		cinfo->tagline = gnome_config_private_get_string("tagline=");
	
		/* Done with this section */
		gnome_config_pop_prefix();
		g_free(secname);

		/* Add it to the clist */
		names[0] = cinfo->name;
		row = gtk_clist_append(GTK_CLIST(cd->clist), names);
		gtk_clist_set_row_data(GTK_CLIST(cd->clist), row, cinfo);
		cd->saved = g_slist_append(cd->saved, cinfo);
	}
	
	/* Done loading */
	gnome_config_pop_prefix();
}

/* connect_dialog_save
 * Take the list of connections from the dialog and save it into the gnome
 * configuration file structure
 */
static void
connect_dialog_save(ConnectDialog *cd)
{
	GSList *list;
	gint    num;
	
	g_return_if_fail(cd);
	
	/* Everything's under here */
	gnome_config_push_prefix("/" APPNAME "/saved-connections/");

	/* Iterate through the connections in the list */
	num = 0;
	for (list = cd->saved; list; list = g_slist_next(list)) {
		ConnectInfo *cinfo;
		gchar       *secname;

		/* Get the data out */
		cinfo = (ConnectInfo *)list->data;
		g_assert(cinfo);

		/* Connections get their own section */
		secname = g_strdup_printf("/%s/saved-connection-%d/",
					  APPNAME, num);
		gnome_config_push_prefix(secname);

		/* Write the data from the structure */
		gnome_config_private_set_string("name", cinfo->name);
		gnome_config_private_set_string("server-name",
						cinfo->server_name);
		gnome_config_private_set_string("server-port",
						cinfo->server_port);
		gnome_config_private_set_string("nick", cinfo->nick);
		gnome_config_private_set_bool("as_guest", cinfo->as_guest);
		gnome_config_private_set_string("server-pass",
						cinfo->server_pass);
		gnome_config_private_set_string("email", cinfo->email);
		gnome_config_private_set_string("tagline", cinfo->tagline);

		/* Done with this section */
		gnome_config_pop_prefix();
		g_free(secname);
		num++;
	}

	/* Write the number of connections out */
	gnome_config_private_set_int("number", num);

	/* Done loading */
	gnome_config_pop_prefix();
	gnome_config_sync();
}

/* connect_dialog_apply_clicked
 * Called when the user clicks the Apply or Ok button.  Save all the current
 * dialog values to disk.
 */
static void
connect_dialog_apply_clicked(GtkWidget *widget, ConnectDialog *cd)
{
	g_return_if_fail(cd);

	/* Save */
	connect_dialog_save(cd);

	/* Saved, so can't apply now */
	gnome_dialog_set_sensitive(GNOME_DIALOG(cd->dialog), 1, FALSE);
}

/* connect_dialog_ok_clicked
 * Called then the user clickes the Ok button.  Connect to the currently
 * selected server if there is one.
 */
static void
connect_dialog_ok_clicked(GtkWidget *widget, ConnectDialog *cd)
{
	ConnectInfo *cinfo;
	Say2Server  *server;

	g_return_if_fail(cd);

	if (!cd->current)
		return;
	cinfo = cd->current;

	/* Make a server connection */
	server = say2_server_new(cinfo->server_name, cinfo->server_port);
	if (!server)
		return;

	/* Set who we login as */
	if (cinfo->as_guest) {
		say2_server_set_guest(server, cinfo->nick, cinfo->email,
				      cinfo->tagline);
	} else {
		say2_server_set_login(server, cinfo->nick, cinfo->server_pass);
	}

	/* Point a ChatInterface at it */
	chat_interface_get(server, CHAT_RECIPTYPE_ANY, NULL);
}

/* connect_dialog_new_clicked
 * Called when the user clicks the New button.  Create a new ConnectInfo called
 * "New Connection" (how original), add it to the clist and list inside the
 * dialog
 */
static void
connect_dialog_new_clicked(GtkWidget *widget, ConnectDialog *cd)
{
	ConnectInfo *cinfo;
	gchar       *names[1];
	gint         row;

	g_return_if_fail(cd);

	/* Create the new ConnectInfo structure, and copy in default values */
	cinfo = g_new0(ConnectInfo, 1);
	cinfo->name = g_strdup(_("New Connection"));
	cinfo->server_name = g_strdup(DEFAULT_SERVER_NAME);
	cinfo->server_port = g_strdup(DEFAULT_SERVER_PORT);
	cinfo->nick = g_strdup("");
	cinfo->server_pass = g_strdup("");
	cinfo->email = g_strdup("");
	cinfo->tagline = g_strdup("");

	/* Add it to the clist */
	names[0] = cinfo->name;
	row = gtk_clist_append(GTK_CLIST(cd->clist), names);
	gtk_clist_set_row_data(GTK_CLIST(cd->clist), row, cinfo);
	cd->saved = g_slist_append(cd->saved, cinfo);

	/* Select the row, this'll generate the "select-row" signal so we
	 * can update everything
	 */
	gtk_clist_select_row(GTK_CLIST(cd->clist), row, 0);
	if (gtk_clist_row_is_visible(GTK_CLIST(cd->clist), row)
	    != GTK_VISIBILITY_FULL)
		gtk_clist_moveto(GTK_CLIST(cd->clist), row, 0, 1.0, 0);

	/* Changes have been made, activate the apply button */
	gnome_dialog_set_sensitive(GNOME_DIALOG(cd->dialog), 1, TRUE);
}

/* connect_dialog_row_selected
 * Called when the user selects a row in the list, or when a new one is
 * appended and selected by meko itself.  Update the interface with the
 * new information.
 */
static void
connect_dialog_row_selected(GtkWidget *clist, gint row, gint column,
			    GdkEventButton *event, ConnectDialog *cd)
{
	ConnectInfo *cinfo;
	
	g_return_if_fail(clist);
	g_return_if_fail(cd);
	g_return_if_fail(clist == cd->clist);

	/* Get the ConnectInfo associated with that entry */
	cinfo = gtk_clist_get_row_data(GTK_CLIST(cd->clist), row);
	g_assert(cinfo);

	/* Select it */
	connect_dialog_select(cd, cinfo);
}

/* connect_dialog_row_unselected
 * Called when the user deselects a row from the list.  Update the interface
 * with disabled and empty information
 */
static void
connect_dialog_row_unselected(GtkWidget *clist, gint row, gint column,
			      GdkEventButton *event, ConnectDialog *cd)
{
	g_return_if_fail(clist);
	g_return_if_fail(cd);
	g_return_if_fail(clist == cd->clist);

	/* Select nothing */
	connect_dialog_select(cd, NULL);
}

/* connect_dialog_select
 * Setup the dialog to display the given ConnectInfo stuff
 */
static void
connect_dialog_select(ConnectDialog *cd, ConnectInfo *cinfo)
{
	g_return_if_fail(cd);

	/* If no cinfo is given, we need to make everything blank.  This also
	 * means that the Remove button shouldn't work anymore
	 */
	if (!cinfo) {
		/* Unset current first so we don't generate events */
		cd->current = NULL;
		gtk_widget_set_sensitive(cd->frame_vbox, FALSE);
		gtk_widget_set_sensitive(cd->remove_button, FALSE);
		gtk_entry_set_text(GTK_ENTRY(cd->conn_name), "");
		gtk_entry_set_text(GTK_ENTRY(cd->server_name), "");
		gtk_entry_set_text(GTK_ENTRY(cd->server_port), "");
		gtk_entry_set_text(GTK_ENTRY(cd->nick), "");
		gtk_entry_set_text(GTK_ENTRY(cd->server_pass), "");
		gtk_entry_set_text(GTK_ENTRY(cd->email), "");
		gtk_entry_set_text(GTK_ENTRY(cd->tagline), "");
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(cd->as_user),
					     TRUE);
		return;
	}

	/* Make the dialog and remove button sensitive, we've got a saved
	 * connection so it can be edited and removed
	 */
	gtk_widget_set_sensitive(cd->frame_vbox, TRUE);
	gtk_widget_set_sensitive(cd->remove_button, TRUE);

	/* Copy the saved text from the cinfo into the entry boxes */
	gtk_entry_set_text(GTK_ENTRY(cd->conn_name), cinfo->name);
	gtk_entry_set_text(GTK_ENTRY(cd->server_name), cinfo->server_name);
	gtk_entry_set_text(GTK_ENTRY(cd->server_port), cinfo->server_port);
	gtk_entry_set_text(GTK_ENTRY(cd->nick), cinfo->nick);
	gtk_entry_set_text(GTK_ENTRY(cd->server_pass), cinfo->server_pass);
	gtk_entry_set_text(GTK_ENTRY(cd->email), cinfo->email);
	gtk_entry_set_text(GTK_ENTRY(cd->tagline), cinfo->tagline);

	/* Toggle the right button */
	if (cinfo->as_guest) {
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(cd->as_guest),
					     TRUE);
	} else {
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(cd->as_user),
					     TRUE);
	}

	/* Do this last so the events don't get generated for the above stuff */
	cd->current = cinfo;
}

/* connect_dialog_remove_clicked
 * Called whenever the user clicks the remove button to remove a saved
 * connection.
 */
static void
connect_dialog_remove_clicked(GtkWidget *widget, ConnectDialog *cd)
{
	ConnectInfo *cinfo;
	gint         row;

	g_return_if_fail(cd);
	g_return_if_fail(cd->current);

	/* Find the row in the list */
	cinfo = cd->current;
	row = gtk_clist_find_row_from_data(GTK_CLIST(cd->clist), cinfo);
	g_return_if_fail(row != -1);

	/* Remove the row from the lists (will call unselected for us) */
	gtk_clist_remove(GTK_CLIST(cd->clist), row);
	cd->saved = g_slist_remove(cd->saved, cinfo);

	/* Free up the info */
	connect_dialog_free_info(cinfo);
}

/* connect_dialog_button_pressed
 * Called when a mouse button is pressed on the list widget.  We use this to
 * check if they double-clicked the left mouse button, and if so treat it as
 * a select and ok in one.
 */
static gint
connect_dialog_button_pressed(GtkWidget *widget, GdkEventButton *event,
			      ConnectDialog *cd)
{
	g_return_val_if_fail(cd, FALSE);

	/* Only capture if there's something selected */
	if (!cd->current)
		return FALSE;
	
	/* Double click on the left mouse button */
	if ((event->type == GDK_2BUTTON_PRESS) && (event->button == 1)) {
		/* We have to do this manually because there's no function
		 * like gnome_dialog_clicked(), which would be really useful.
		 * sigh.
		 */
		connect_dialog_apply_clicked(widget, cd);
		connect_dialog_ok_clicked(widget, cd);
		gtk_widget_destroy(cd->dialog);
		return TRUE;
	}

	return FALSE;
}

/* connect_dialog_entry_changed
 * Called whenever the user changes one of the text entries on the interface.
 * Updates the information in the ConnectInfo structure (and in the case of
 * the connection name, the list title too).
 */
static void
connect_dialog_entry_changed(GtkWidget *entry, ConnectDialog *cd)
{
	const gchar *text;
	ConnectInfo *cinfo;

	g_return_if_fail(entry);
	g_return_if_fail(cd);

	/* cd->current will be NULL when switching between rows.  This prevents
	 * us doing all this when changing info.  It's not a problem if we do,
	 * it's just work we don't need to do.
	 */
	if (!cd->current)
		return;

	/* Get the text off the widget */
	text = gtk_entry_get_text(GTK_ENTRY(entry));
	cinfo = cd->current;

	if (entry == cd->conn_name) {
		/* Connection name was changed.  Update the list title too */
		gint row;

		/* Update the structure */
		g_free(cinfo->name);
		cinfo->name = g_strdup(text);

		/* Get the current row */
		row = gtk_clist_find_row_from_data(GTK_CLIST(cd->clist),
						   cd->current);
		g_return_if_fail(row != -1);

		/* Set the title of the row */
		gtk_clist_set_text(GTK_CLIST(cd->clist), row, 0, text);

	} else if (entry == cd->server_name) {
		/* Server Name was changed. */
		g_free(cinfo->server_name);
		cinfo->server_name = g_strdup(text);

	} else if (entry == cd->server_port) {
		/* Server Port was changed. */
		g_free(cinfo->server_port);
		cinfo->server_port = g_strdup(text);

	} else if (entry == cd->nick) {
		/* Nickname was changed. */
		g_free(cinfo->nick);
		cinfo->nick = g_strdup(text);

	} else if (entry == cd->server_pass) {
		/* Password was changed. */
		g_free(cinfo->server_pass);
		cinfo->server_pass = g_strdup(text);

	} else if (entry == cd->email) {
		/* E-Mail Address was changed. */
		g_free(cinfo->email);
		cinfo->email = g_strdup(text);

	} else if (entry == cd->tagline) {
		/* Tagline was changed. */
		g_free(cinfo->tagline);
		cinfo->tagline = g_strdup(text);

	} else {
		g_assert_not_reached();
	}

	/* Changes have been made, activate the apply button */
	gnome_dialog_set_sensitive(GNOME_DIALOG(cd->dialog), 1, TRUE);
}

/* connect_dialog_login_toggled
 * Called whenever the user toggles either of the "Login as" radio buttons.
 * GTK+ will actually call this once for each member of the group, so we can
 * neaten the code by hooking it the right way.
 */
static void
connect_dialog_login_toggled(GtkWidget *widget, ConnectDialog *cd)
{
	gboolean active;

	g_return_if_fail(cd);

	active = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));

	if (widget == cd->as_user) {
		gtk_widget_set_sensitive(cd->server_pass_label, active);
		gtk_widget_set_sensitive(cd->server_pass, active);
	} else if (widget == cd->as_guest) {
		gtk_widget_set_sensitive(cd->email_label, active);
		gtk_widget_set_sensitive(cd->email, active);
		gtk_widget_set_sensitive(cd->tagline_label, active);
		gtk_widget_set_sensitive(cd->tagline, active);

		/* We only need to hook as_guest for this, because the event
		 * is always generated for both radios and this one just makes
		 * more sense.
		 */
		if (cd->current) {
			cd->current->as_guest = active;

			/* Changes have been made, activate the apply button */
			gnome_dialog_set_sensitive(GNOME_DIALOG(cd->dialog),
						   1, TRUE);
		}
	}
}

/* connect_dialog_destroy
 * Called when the dialog is destroyed.  We get passed a pointer to a pointer
 * because we need to free it and set it to 0 (it's a static)
 */
static void
connect_dialog_destroy(GtkWidget *widget, ConnectDialog **cd)
{
	GSList *list;

	g_return_if_fail(cd);

	if (!(*cd))
		return;

	/* Iterate the saved info, freeing it up */
	for (list = (*cd)->saved; list; list = g_slist_next(list)) {
		connect_dialog_free_info((ConnectInfo *)list->data);
		list->data = NULL;
	}
	g_slist_free((*cd)->saved);

	/* Free the dialog, structure and set the pointer to NULL */
	gtk_widget_destroy((*cd)->dialog);
	g_free(*cd);
	*cd = NULL;
}

/* connect_dialog_free_info
 * Free the contents of a ConnectInfo structure
 */
static void
connect_dialog_free_info(ConnectInfo *cinfo)
{
	g_free(cinfo->name);
	g_free(cinfo->server_name);
	g_free(cinfo->server_port);
	g_free(cinfo->nick);
	g_free(cinfo->server_pass);
	g_free(cinfo->email);
	g_free(cinfo->tagline);
	g_free(cinfo);
}

/* connect_dialog_new
 * Create a new connection dialog.  This won't show it or run it, the caller
 * will need to do that.
 */
GtkWidget *
connect_dialog_new(GtkWidget *parent)
{
	static ConnectDialog *cd = NULL;
	GtkWidget            *top_hbox, *list_vbox, *scrollwin, *list_hbox;
	GtkWidget            *button, *frame, *frame_vbox, *label_frame, *label;
	GtkWidget            *hbox, *table, *separator;
	GSList		     *group;

	/* If we've already made one then just reshow it */
	if (cd) {
		gdk_window_show(cd->dialog->window);
		gdk_window_raise(cd->dialog->window);
		return cd->dialog;
	}

	/* Create the structure */
	cd = g_new0(ConnectDialog, 1);

	/* Create the dialog with the standard preferences dialog buttons */
	cd->dialog = gnome_dialog_new(_("Meko: Connection"),
				      GNOME_STOCK_BUTTON_OK,
				      GNOME_STOCK_BUTTON_APPLY,
				      GNOME_STOCK_BUTTON_CLOSE,
				      GNOME_STOCK_BUTTON_HELP,
				      NULL);
	gtk_window_set_wmclass(GTK_WINDOW(cd->dialog), "connect_dialog",
			       APPNAME);
	gnome_dialog_set_default(GNOME_DIALOG(cd->dialog), 0);
	gnome_dialog_set_sensitive(GNOME_DIALOG(cd->dialog), 1, FALSE);
	gnome_dialog_set_sensitive(GNOME_DIALOG(cd->dialog), 3, FALSE);
	gtk_signal_connect(GTK_OBJECT(cd->dialog), "destroy",
			   GTK_SIGNAL_FUNC(connect_dialog_destroy),
			   &cd);

	/* Connect signal handlers to the Ok button.  It basically calls
	 * Apply, then Ok and finally gtk_widget_destroy.
	 */
	gnome_dialog_button_connect
		(GNOME_DIALOG(cd->dialog), 0,
		 GTK_SIGNAL_FUNC(connect_dialog_apply_clicked), (gpointer)cd);
	gnome_dialog_button_connect
		(GNOME_DIALOG(cd->dialog), 0,
		 GTK_SIGNAL_FUNC(connect_dialog_ok_clicked), (gpointer)cd);
	gnome_dialog_button_connect_object
		(GNOME_DIALOG(cd->dialog), 0,
		 GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(cd->dialog));

	/* Connect signal handlers to the Apply button.  Just call Apply */
	gnome_dialog_button_connect
		(GNOME_DIALOG(cd->dialog), 1,
		 GTK_SIGNAL_FUNC(connect_dialog_apply_clicked), (gpointer)cd);

	/* Connect signal handlers to the Close button.  Just calls
	 * gtk_widget_destroy (ie doesn't save).
	 */
	gnome_dialog_button_connect_object
		(GNOME_DIALOG(cd->dialog), 2,
		 GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(cd->dialog));

	/* A good dialog should have a good parent */
	if (parent)
		gnome_dialog_set_parent(GNOME_DIALOG(cd->dialog),
					GTK_WINDOW(parent));

	/* Horizontal packing box, this is for the two halves of the dialog,
	 * on the left the saved connections list and on the right the
	 * Frame'O'Stuff(tm)
	 */
	top_hbox = gtk_hbox_new(FALSE, 6);
	gtk_box_pack_start(GTK_BOX(GNOME_DIALOG(cd->dialog)->vbox), top_hbox,
			   TRUE, TRUE, 0);
	gtk_widget_show(top_hbox);


	/* On the left we've got a list of settings we've saved, with a
	 * couple of buttons below to create new settings and to remove
	 * highlighted ones
	 */
	list_vbox = gtk_vbox_new(FALSE, 2);
	gtk_box_pack_start(GTK_BOX(top_hbox), list_vbox, FALSE, FALSE, 0);
	gtk_widget_show(list_vbox);

	/* Scrolling window for the list */
	scrollwin = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrollwin),
				       GTK_POLICY_AUTOMATIC,
				       GTK_POLICY_AUTOMATIC);
	gtk_box_pack_start(GTK_BOX(list_vbox), scrollwin, TRUE, TRUE, 0);
	gtk_widget_show(scrollwin);

	/* Create the list */
	cd->clist = gtk_clist_new(1);
	gtk_widget_set_usize(cd->clist, 145, -1);
	gtk_clist_set_column_title(GTK_CLIST(cd->clist), 0,
				   _("Saved Connections"));
	gtk_clist_column_titles_show(GTK_CLIST(cd->clist));
	gtk_clist_column_titles_passive(GTK_CLIST(cd->clist));
	gtk_clist_set_column_auto_resize(GTK_CLIST(cd->clist), 0, TRUE);
	gtk_clist_set_selection_mode(GTK_CLIST(cd->clist),
				     GTK_SELECTION_SINGLE);
	gtk_signal_connect(GTK_OBJECT(cd->clist), "select-row",
			   GTK_SIGNAL_FUNC(connect_dialog_row_selected),
			   (gpointer)cd);
	gtk_signal_connect(GTK_OBJECT(cd->clist), "unselect-row",
			   GTK_SIGNAL_FUNC(connect_dialog_row_unselected),
			   (gpointer)cd);
	gtk_signal_connect(GTK_OBJECT(cd->clist), "button-press-event",
			   GTK_SIGNAL_FUNC(connect_dialog_button_pressed),
			   (gpointer)cd);
	gtk_container_add(GTK_CONTAINER(scrollwin), cd->clist);
	gtk_widget_show(cd->clist);

	/* Need two buttons below, so pack them into a horizontal box */
	list_hbox = gtk_hbox_new(TRUE, 0);
	gtk_box_pack_end(GTK_BOX(list_vbox), list_hbox, FALSE, FALSE, 0);
	gtk_widget_show(list_hbox);

	/* Create a "New" button.  Clicking this will add a "New Connection"
	 * to the list.
	 */
	button = gtk_button_new_with_label(_("New"));
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
			   GTK_SIGNAL_FUNC(connect_dialog_new_clicked),
			   (gpointer)cd);
	gtk_box_pack_start(GTK_BOX(list_hbox), button, TRUE, TRUE, 0);
	gtk_widget_show(button);

	/* Create a "Remove" button.  Clicking this will remove the currently
	 * selected selection.
	 */
	cd->remove_button = gtk_button_new_with_label(_("Remove"));
	gtk_widget_set_sensitive(cd->remove_button, FALSE);
	gtk_signal_connect(GTK_OBJECT(cd->remove_button), "clicked",
			   GTK_SIGNAL_FUNC(connect_dialog_remove_clicked),
			   (gpointer)cd);
	gtk_box_pack_start(GTK_BOX(list_hbox), cd->remove_button,
			   TRUE, TRUE, 0);
	gtk_widget_show(cd->remove_button);


	/* On the right we have a pretty frame containing the rest */
	frame = gtk_frame_new(NULL);
	gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_IN);
	gtk_box_pack_end(GTK_BOX(top_hbox), frame, TRUE, TRUE, 0);
	gtk_widget_show(frame);

	/* Inside the frame we need a vbox, so we can pack the frame title
	 * and the top and everything else inside
	 */
	frame_vbox = gtk_vbox_new(FALSE, 4);
	gtk_container_add(GTK_CONTAINER(frame), frame_vbox);
	gtk_widget_show(frame_vbox);

	/* We want to make the text look pretty, so we'll place a frame
	 * around it etched out
	 */
	label_frame = gtk_frame_new(NULL);
	gtk_frame_set_shadow_type(GTK_FRAME(label_frame), GTK_SHADOW_OUT);
	gtk_box_pack_start(GTK_BOX(frame_vbox), label_frame, FALSE, FALSE, 0);
	gtk_widget_show(label_frame);

	/* Now put the label in the frame, aligned it to the left */
	label = gtk_label_new(_("Connection Settings"));
	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
	gtk_container_add(GTK_CONTAINER(label_frame), label);
	gtk_widget_show(label);


	/* Packing box for everything in the frame */
	cd->frame_vbox = gtk_vbox_new(FALSE, 2);
	gtk_widget_set_sensitive(cd->frame_vbox, FALSE);
	gtk_container_set_border_width(GTK_CONTAINER(cd->frame_vbox), 4);
	gtk_box_pack_end(GTK_BOX(frame_vbox), cd->frame_vbox, TRUE, TRUE, 0);
	gtk_widget_show(cd->frame_vbox);


	/* Packing box */
	hbox = gtk_hbox_new(FALSE, 4);
	gtk_box_pack_start(GTK_BOX(cd->frame_vbox), hbox, FALSE, FALSE, 0);
	gtk_widget_show(hbox);

	/* Label */
	label = gtk_label_new(_("Connection Name:"));
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	gtk_widget_show(label);

	/* Connection Name entry box */
	cd->conn_name = gtk_entry_new();
	gtk_signal_connect(GTK_OBJECT(cd->conn_name), "changed",
			   GTK_SIGNAL_FUNC(connect_dialog_entry_changed),
			   (gpointer)cd);
	gtk_box_pack_end(GTK_BOX(hbox), cd->conn_name, TRUE, TRUE, 0);
	gtk_widget_show(cd->conn_name);


	/* Frame for server and port */
	frame = gtk_frame_new(_("Connect To:"));
	gtk_box_pack_start(GTK_BOX(cd->frame_vbox), frame, FALSE, FALSE, 0);
	gtk_widget_show(frame);


	/* Packing box */
	hbox = gtk_hbox_new(FALSE, 4);
	gtk_container_set_border_width(GTK_CONTAINER(hbox), 2);
	gtk_container_add(GTK_CONTAINER(frame), hbox);
	gtk_widget_show(hbox);

	/* Label */
	label = gtk_label_new(_("Server:"));
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	gtk_widget_show(label);

	/* Server Name entry box */
	cd->server_name = gtk_entry_new();
	gtk_signal_connect(GTK_OBJECT(cd->server_name), "changed",
			   GTK_SIGNAL_FUNC(connect_dialog_entry_changed),
			   (gpointer)cd);
	gtk_box_pack_start(GTK_BOX(hbox), cd->server_name, TRUE, TRUE, 0);
	gtk_widget_show(cd->server_name);

	/* Label */
	label = gtk_label_new(_("Port:"));
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	gtk_widget_show(label);

	/* Server Port entry box */
	cd->server_port = gtk_entry_new();
	gtk_widget_set_usize(cd->server_port, 50, -1);
	gtk_signal_connect(GTK_OBJECT(cd->server_port), "changed",
			   GTK_SIGNAL_FUNC(connect_dialog_entry_changed),
			   (gpointer)cd);
	gtk_box_pack_start(GTK_BOX(hbox), cd->server_port, FALSE, FALSE, 0);
	gtk_widget_show(cd->server_port);


	/* Frame for login as stuff */
	frame = gtk_frame_new(_("Connect As:"));
	gtk_box_pack_start(GTK_BOX(cd->frame_vbox), frame, FALSE, FALSE, 0);
	gtk_widget_show(frame);

	/* Table for everything in the frame */
	table = gtk_table_new(6, 2, FALSE);
	gtk_container_set_border_width(GTK_CONTAINER(table), 2);
	gtk_container_add(GTK_CONTAINER(frame), table);
	gtk_widget_show(table);


	/* Packing box */
	hbox = gtk_hbox_new(FALSE, 0);
	gtk_table_attach(GTK_TABLE(table), hbox, 0, 2, 0, 1,
			 GTK_SHRINK | GTK_FILL, GTK_SHRINK | GTK_FILL, 0, 0);
	gtk_widget_show(hbox);

	/* Label */
	label = gtk_label_new(_("Nickname:"));
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	gtk_widget_show(label);

	/* Nickname entry box */
	cd->nick = gtk_entry_new();
	gtk_signal_connect(GTK_OBJECT(cd->nick), "changed",
			   GTK_SIGNAL_FUNC(connect_dialog_entry_changed),
			   (gpointer)cd);
	gtk_box_pack_start(GTK_BOX(hbox), cd->nick, TRUE, TRUE, 0);
	gtk_widget_show(cd->nick);


	/* Separator */
	separator = gtk_hseparator_new();
	gtk_table_attach(GTK_TABLE(table), separator, 0, 2, 1, 2,
			 GTK_EXPAND | GTK_FILL, GTK_SHRINK | GTK_FILL, 0, 4);
	gtk_widget_show(separator);


	/* Login as registered user button */
	cd->as_user = gtk_radio_button_new_with_label
		(NULL, _("Login as Registered User"));
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(cd->as_user));
	gtk_signal_connect(GTK_OBJECT(cd->as_user), "toggled",
			   GTK_SIGNAL_FUNC(connect_dialog_login_toggled),
			   (gpointer)cd);
	gtk_table_attach(GTK_TABLE(table), cd->as_user, 0, 2, 2, 3,
			 GTK_SHRINK | GTK_FILL, GTK_SHRINK | GTK_FILL, 0, 0);
	gtk_widget_show(cd->as_user);

	/* Login as guest button */
	cd->as_guest = gtk_radio_button_new_with_label
		(group, _("Login as Guest"));
	gtk_signal_connect(GTK_OBJECT(cd->as_guest), "toggled",
			   GTK_SIGNAL_FUNC(connect_dialog_login_toggled),
			   (gpointer)cd);
	gtk_table_attach(GTK_TABLE(table), cd->as_guest, 0, 2, 4, 5,
			 GTK_SHRINK | GTK_FILL, GTK_SHRINK | GTK_FILL, 0, 0);
	gtk_widget_show(cd->as_guest);


	/* Label */
	cd->server_pass_label = gtk_label_new(_("Password:"));
	gtk_misc_set_alignment(GTK_MISC(cd->server_pass_label), 1.0, 0.5);
	gtk_table_attach(GTK_TABLE(table), cd->server_pass_label, 0, 1, 3, 4,
			 GTK_SHRINK | GTK_FILL, GTK_SHRINK | GTK_FILL, 0, 0);
	gtk_widget_show(cd->server_pass_label);

	/* Password entry box */
	cd->server_pass = gtk_entry_new();
	gtk_entry_set_visibility(GTK_ENTRY(cd->server_pass), FALSE);
	gtk_signal_connect(GTK_OBJECT(cd->server_pass), "changed",
			   GTK_SIGNAL_FUNC(connect_dialog_entry_changed),
			   (gpointer)cd);
	gtk_table_attach(GTK_TABLE(table), cd->server_pass, 1, 2, 3, 4,
			 GTK_SHRINK | GTK_FILL, GTK_SHRINK | GTK_FILL, 0, 0);
	gtk_widget_show(cd->server_pass);


	/* Label */
	cd->email_label = gtk_label_new(_("E-Mail:"));
	gtk_widget_set_sensitive(cd->email_label, FALSE);
	gtk_misc_set_alignment(GTK_MISC(cd->email_label), 1.0, 0.5);
	gtk_table_attach(GTK_TABLE(table), cd->email_label, 0, 1, 5, 6,
			 GTK_SHRINK | GTK_FILL, GTK_SHRINK | GTK_FILL, 0, 0);
	gtk_widget_show(cd->email_label);

	/* E-Mail Address entry box */
	cd->email = gtk_entry_new();
	gtk_widget_set_sensitive(cd->email, FALSE);
	gtk_signal_connect(GTK_OBJECT(cd->email), "changed",
			   GTK_SIGNAL_FUNC(connect_dialog_entry_changed),
			   (gpointer)cd);
	gtk_table_attach(GTK_TABLE(table), cd->email, 1, 2, 5, 6,
			 GTK_SHRINK | GTK_FILL, GTK_SHRINK | GTK_FILL, 0, 0);
	gtk_widget_show(cd->email);


	/* Label */
	cd->tagline_label = gtk_label_new(_("Tag Line:"));
	gtk_widget_set_sensitive(cd->tagline_label, FALSE);
	gtk_misc_set_alignment(GTK_MISC(cd->tagline_label), 1.0, 0.5);
	gtk_table_attach(GTK_TABLE(table), cd->tagline_label, 0, 1, 6, 7,
			 GTK_SHRINK | GTK_FILL, GTK_SHRINK | GTK_FILL, 0, 0);
	gtk_widget_show(cd->tagline_label);

	/* Tag Line entry box */
	cd->tagline = gtk_entry_new();
	gtk_widget_set_sensitive(cd->tagline, FALSE);
	gtk_signal_connect(GTK_OBJECT(cd->tagline), "changed",
			   GTK_SIGNAL_FUNC(connect_dialog_entry_changed),
			   (gpointer)cd);
	gtk_table_attach(GTK_TABLE(table), cd->tagline, 1, 2, 6, 7,
			 GTK_SHRINK | GTK_FILL, GTK_SHRINK | GTK_FILL, 0, 0);
	gtk_widget_show(cd->tagline);


	/* Load any saved settings */
	connect_dialog_load(cd);

	/* Return the dialog to be saved or shown etc */
	return cd->dialog;
}
