/* meko
 * Copyright (C) 2001 Scott James Remnant <scott@netsplit.com>.
 * All Rights Reserved.
 *
 * meko_app.c
 *  - Main application window
 * -- 
 * $Id: meko_app.c,v 1.8 2001/06/17 18:27:08 scott Exp $
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <meko.h>

#include <glib.h>
#include <gnome.h>
#include <gtk/gtk.h>
#include <libgnomeui/gnome-window-icon.h>

#include "meko_app.h"
#include "chat_interface.h"
#include "connect_dialog.h"

/* Forward prototypes for callbacks */
static void meko_app_close(GtkWidget *, GtkWidget *);
static void meko_app_destroy(GtkWidget *, gpointer);
static void meko_app_connect(GtkWidget *, GtkWidget *);
static void meko_app_disconnect(GtkWidget *, GtkWidget *);
static void meko_app_about(GtkWidget *, GtkWidget *);
static void meko_app_notyet(GtkWidget *, GtkWidget *);

/* File menu */
static GnomeUIInfo file_menu[] = {
	GNOMEUIINFO_MENU_NEW_WINDOW_ITEM(meko_app_notyet, NULL),
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_ITEM_STOCK(N_("_Connect..."),
			       N_("Connect to a Say2 chat server"),
			       meko_app_connect, GNOME_STOCK_PIXMAP_EXEC),
	GNOMEUIINFO_ITEM_STOCK(N_("_Disconnect"),
			       N_("Disconnect from the current Say2 server"),
			       meko_app_disconnect, GNOME_STOCK_PIXMAP_STOP),
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_MENU_CLOSE_ITEM(meko_app_close, NULL),
	GNOMEUIINFO_MENU_EXIT_ITEM(meko_app_destroy, NULL),
	GNOMEUIINFO_END
};

/* Settings menu */
static GnomeUIInfo settings_menu[] = {
	GNOMEUIINFO_MENU_PREFERENCES_ITEM(meko_app_notyet, NULL),
	GNOMEUIINFO_END
};

/* Help menu */
static GnomeUIInfo help_menu[] = {
	GNOMEUIINFO_HELP(PACKAGE),
	GNOMEUIINFO_MENU_ABOUT_ITEM(meko_app_about, NULL),
	GNOMEUIINFO_END
};

/* Menu bar */
static GnomeUIInfo menu_bar[] = {
	GNOMEUIINFO_MENU_FILE_TREE(file_menu),
	GNOMEUIINFO_MENU_SETTINGS_TREE(settings_menu),
	GNOMEUIINFO_MENU_HELP_TREE(help_menu),
	GNOMEUIINFO_END
};


/* meko_app_show
 * Show the application window
 */
void
meko_app_show(GtkWidget *app)
{
	ChatInterface *ci;

	g_return_if_fail(app);

	/* Get the user data off the app (it's the ChatInterface inside it) */
	ci = (ChatInterface *)gtk_object_get_data(GTK_OBJECT(app),
						  "chat_interface");
	g_assert(ci);

	/* Show the app */
	gtk_widget_show(app);

	/* Focus the entry box */
	gtk_widget_grab_focus(ci->entry_box);
}

/* meko_app_close
 * Close the parent window of this menu widget
 */
static void
meko_app_close(GtkWidget *widget, GtkWidget *win)
{
	g_assert(win);

	/* Close the window */
	gtk_widget_destroy(win);
}

/* meko_app_connect
 * Called when the user selects connect from the menu.  Disconnect the chat
 * interface from its server and throw up a ConnectDialog to specify a new
 * one
 */
static void
meko_app_connect(GtkWidget *widget, GtkWidget *win)
{
	ChatInterface *ci;
	GtkWidget     *dialog;

	g_return_if_fail(win);

	/* Get the ChatInterface */
	ci = gtk_object_get_data(GTK_OBJECT(win), "chat_interface");
	g_assert(ci);

	/* Disconnect it */
	chat_interface_disconnect(ci);

	/* Open a ConnectDialog */
	dialog = connect_dialog_new(win);
	gtk_widget_show(dialog);
}

/* meko_app_disconnect
 * Called when the user selects disconnect from the menu.  Close all chat
 * interfaces with the server on the current one.
 */
static void
meko_app_disconnect(GtkWidget *widget, GtkWidget *win)
{
	ChatInterface *ci;

	g_return_if_fail(win);

	/* Get the ChatInterface */
	ci = gtk_object_get_data(GTK_OBJECT(win), "chat_interface");
	g_assert(ci);

	/* Disconnect them */
	if (ci->server)
		chat_interface_update(ci->server, CHAT_UPDATE_CLOSE, NULL);
}

/* meko_app_destroy
 * Called when the main window receives a destroy signal, or when the user
 * chooses to quit.  Leave gtk_main and therefore terminate the program.
 */
static void
meko_app_destroy(GtkWidget *widget, gpointer data)
{
	gtk_main_quit();
}

/* meko_app_about
 * Show the about window (yay!)
 */
static void
meko_app_about(GtkWidget *widget, GtkWidget *win)
{
	const gchar      *authors[] = {
		"Scott James Remnant <scott@netsplit.com>",
		"Robey Pointer <robey@lag.net> (Say2 Library)",
		NULL
	};
	
	static GtkWidget *about = NULL;
	GtkWidget        *hbox, *href;

	g_assert(win);

	if (about) {
		/* The dialog is already open somewhere, so just show and
		 * raise it
		 */
		gdk_window_show(about->window);
		gdk_window_raise(about->window);
		return;
	}

	/* Create the about dialog */
	about = gnome_about_new(APPNAME, VERSION,
				"(C) 2001 Scott James Remnant", authors,
				_("The GNOME Say2 chat client"),
				GNOMEPIXMAPDIR "/" PACKAGE ".png");
	gtk_window_set_wmclass(GTK_WINDOW(about), "about_dialog", APPNAME);
	gnome_dialog_set_parent(GNOME_DIALOG(about), GTK_WINDOW(win));
	gtk_signal_connect(GTK_OBJECT(about), "destroy",
			   GTK_SIGNAL_FUNC(gtk_widget_destroyed),
			   &about);

	/* Create an hbox in it to put an href button in */
	hbox = gtk_hbox_new(TRUE, 0);
	gtk_box_pack_start(GTK_BOX(GNOME_DIALOG(about)->vbox), hbox,
			   FALSE, FALSE, 0);
	gtk_widget_show(hbox);

	/* Put an href button in */
	href = gnome_href_new("http://www.lag.net/say2/",
			      _("The Say2 Homepage"));
	gtk_box_pack_start(GTK_BOX(hbox), href, FALSE, FALSE, 0);
	gtk_widget_show(href);

	/* Show the about dialog */
	gtk_widget_show(about);
}

/* meko_app_notyet
 * Placeholder function for menu entries we don't handle yet
 */
static void
meko_app_notyet(GtkWidget *widget, GtkWidget *win)
{
	g_return_if_fail(win);
	
	gnome_app_message(GNOME_APP(win),
			  _("Sorry, that menu item is not yet implemented."));
}


/* meko_app_new
 * Create the application main window.
 */
GtkWidget *
meko_app_new(const gchar *geometry)
{
	GtkWidget     *app, *appbar;
	ChatInterface *ci;

	/* Create the GnomeApp */
	app = gnome_app_new(PACKAGE, APPNAME " " FULL_VERSION);
	gtk_window_set_wmclass(GTK_WINDOW(app), "main_window", APPNAME);
	gtk_window_set_policy(GTK_WINDOW(app), TRUE, TRUE, FALSE);
	gtk_window_set_default_size(GTK_WINDOW(app), 576, 352);
	gtk_signal_connect(GTK_OBJECT(app), "destroy",
			   GTK_SIGNAL_FUNC(meko_app_destroy), NULL);

	/* If the user specified a geometry, parse and set it */
	if (geometry) {
		gint x, y, w, h;

		if (gnome_parse_geometry(geometry, &x, &y, &w, &h)) {
			if (x != -1)
				gtk_widget_set_uposition(app, x, y);
			if (w != -1)
				gtk_window_set_default_size(GTK_WINDOW(app),
							    w, h);
		} else {
			g_error("%s `%s'\n",
				_("Could not parse geometry string"), geometry);
		}
	}

	/* Set the icon */
	gnome_window_icon_set_from_file(GTK_WINDOW(app),
					GNOMEPIXMAPDIR "/" PACKAGE ".png");

	/* Create an app bar (status bar) */
	appbar = gnome_appbar_new(FALSE, TRUE, GNOME_PREFERENCES_USER);
	gnome_app_set_statusbar(GNOME_APP(app), appbar);
	gtk_widget_show(appbar);

	/* Attach the menu bar */
	gnome_app_create_menus_with_data(GNOME_APP(app), menu_bar, app);
	gnome_app_install_menu_hints(GNOME_APP(app), menu_bar);

	/* Create a Chat Interface */
	ci = chat_interface_new();
	gtk_object_set_data(GTK_OBJECT(app), "chat_interface", ci);
	gtk_container_set_border_width(GTK_CONTAINER(ci->top_vbox), 2);
	gnome_app_set_contents(GNOME_APP(app), ci->top_vbox);

	/* Return the app */
	return app;
}
