/* meko
 * Copyright (C) 2001 Scott James Remnant <scott@netsplit.com>.
 * All Rights Reserved.
 *
 * say2_client.c
 *  - Handles the Say2Client structure which contains information about
 *    users we want to know about, including ourselves
 * -- 
 * $Id: say2_client.c,v 1.5 2001/06/17 17:53:51 scott Exp $
 * 
 * This file is distributed according to the GNU General Public
 * License.  For full details, read the top of 'main.c' or the
 * file called COPYING that was distributed with this code.
 */

#include <meko.h>

#include <glib.h>

#include "say2_types.h"
#include "say2_client.h"
#include "whois_info.h"


/* say2_client_copy_from_whois
 * Copy data from a WhoisInfo structure into a Say2Client structure
 */
void
say2_client_copy_from_whois(Say2Client *client, const WhoisInfo *wi)
{
	g_return_if_fail(client);
	g_return_if_fail(wi);

	/* Copy the string values */
	if (wi->id) {
		g_free(client->id);
		client->id = g_strdup(wi->id);
	}
	if (wi->nick) {
		g_free(client->nick);
		client->nick = g_strdup(wi->nick);
	}
	if (wi->email) {
		g_free(client->email);
		client->email = g_strdup(wi->email);
	}
	if (wi->tagline) {
		g_free(client->tagline);
		client->tagline = g_strdup(wi->tagline);
	}
	if (wi->status_msg) {
		g_free(client->status_msg);
		client->status_msg = g_strdup(wi->status_msg);
	}

	/* Copy the mode, converting it into a Say2Status instead */
	if (wi->status_mode) {
		if (!g_strcasecmp(wi->status_mode, "active")) {
			client->status = SAY2_STATUS_ACTIVE;
		} else if (!g_strcasecmp(wi->status_mode, "busy")) {
			client->status = SAY2_STATUS_BUSY;
		} else if (!g_strcasecmp(wi->status_mode, "away")) {
			client->status = SAY2_STATUS_AWAY;
		} else {
			client->status = SAY2_STATUS_OTHER;
		}
	}
}

/* say2_client_new_from_whois
 * Allocate room for a new Say2Client struct filling it with values taken
 * from a WhoisInfo struct
 */
Say2Client *
say2_client_new_from_whois(const WhoisInfo *wi)
{
	Say2Client *client;

	client = say2_client_new();
	say2_client_copy_from_whois(client, wi);

	return client;
}

/* say2_client_new
 * Allocate room for a new Say2Client struct
 */
Say2Client *
say2_client_new(void)
{
	return g_new0(Say2Client, 1);
}

/* say2_client_ref
 * Increment the reference count of a Say2Client structure
 */
void
say2_client_ref(Say2Client *client)
{
	if (!client)
		return;

	client->ref_count++;
}

/* say2_client_unref
 * Decrement the reference count of a Say2Client structure, freeing it if
 * the reference count reaches 0
 */
void
say2_client_unref(Say2Client *client)
{
	if (!client)
		return;

	g_assert(client->ref_count);

	client->ref_count--;

	/* Reference count reached 0, free it */
	if (!client->ref_count) {
		g_free(client->id);
		g_free(client->nick);
		g_free(client->email);
		g_free(client->tagline);
		g_free(client->status_msg);
		if (client->icon != (char *)-1)
			g_free(client->icon);
		g_free(client);
	}
}
