/* meko
 * Copyright (C) 2001 Scott James Remnant <scott@netsplit.com>.
 * All Rights Reserved.
 *
 * say2_commands.c
 *  - Handles commands received from a say2 server
 * -- 
 * $Id: say2_commands.c,v 1.8 2001/06/17 16:22:06 scott Exp $
 * 
 * This file is distributed according to the GNU General Public
 * License.  For full details, read the top of 'main.c' or the
 * file called COPYING that was distributed with this code.
 */

#include <meko.h>

#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <glib.h>
#include <gnome.h>
#include <say2/say2.h>

#include "say2_commands.h"
#include "say2_types.h"
#include "say2_server.h"
#include "chat_interface.h"
#include "whois_info.h"
#include "unicode.h"

/* Forward prototypes for command functions*/
static void say2_error(Say2Server *, const EnamlBlock *);
static void say2_hello(Say2Server *, const EnamlBlock *);
static void say2_server_info(Say2Server *, const EnamlBlock *);
static void say2_motd(Say2Server *, const EnamlBlock *);
static void say2_welcome(Say2Server *, const EnamlBlock *);
static void say2_bad_nick(Say2Server *, const EnamlBlock *);
static void say2_ping(Say2Server *, const EnamlBlock *);
static void say2_signon(Say2Server *, const EnamlBlock *);
static void say2_signoff(Say2Server *, const EnamlBlock *);
static void say2_prefs(Say2Server *, const EnamlBlock *);
static void say2_status(Say2Server *, const EnamlBlock *);
static void say2_change_nick(Say2Server *, const EnamlBlock *);
static void say2_members(Say2Server *, const EnamlBlock *);
static void say2_begin_response(Say2Server *, const EnamlBlock *);
static void say2_end_response(Say2Server *, const EnamlBlock *);
static void say2_client_info(Say2Server *, const EnamlBlock *);
static void say2_not_present(Say2Server *, const EnamlBlock *);
static void say2_no_such_client(Say2Server *, const EnamlBlock *);
#if 0
static void say2_client_attributes(Say2Server *, const EnamlBlock *);
static void say2_set_attributes(Say2Server *, const EnamlBlock *);
#endif
static void say2_say(Say2Server *, const EnamlBlock *);
static void say2_message(Say2Server *, const EnamlBlock *);
static void say2_wall(Say2Server *, const EnamlBlock *);

/* Command dispatch table */
static DispatchItem command_table[] = {
	{ "error", &say2_error },
	{ "hello", &say2_hello },
	{ "server-info", &say2_server_info },
	{ "motd", &say2_motd },
	{ "welcome", &say2_welcome },
	{ "bad-nick", &say2_bad_nick },
	{ "ping", &say2_ping },
	{ "signon", &say2_signon },
	{ "signoff", &say2_signoff },
	{ "prefs", &say2_prefs },
	{ "status", &say2_status },
	{ "change-nick", &say2_change_nick },
	{ "members", &say2_members },
	{ "begin-response", &say2_begin_response },
	{ "end-response", &say2_end_response },
	{ "client-info", &say2_client_info },
	{ "not-present", &say2_not_present },
	{ "no-such-client", &say2_no_such_client },
#if 0
	{ "client-attributes", &say2_client_attributes },
	{ "set-attributes", &say2_set_attributes },
#endif
	{ "say", &say2_say },
	{ "message", &say2_message },
	{ "wall", &say2_wall },
	{ NULL, NULL }
};


/* say2_error
 * Handle error command
 */
static void
say2_error(Say2Server *server, const EnamlBlock *error)
{
	const gchar *command;
	gchar       *text;

	g_return_if_fail(server);

	/* Cope with a plain error flag */
	if (!error) {
		chat_interface_say(server, "*** %s\n", _("Error"));
		return;
	}

	/* Get a command and utf-8 text from the block */
	command = enaml_get_string(error, "command", NULL);
	text = utf8_decode(enaml_get_string(error, "text", ""));

	/* Text is optional */
	if (command) {
		chat_interface_say(server, "*** %s %s: %s\n",
				   command, _("Error"), text);
	} else {
		chat_interface_say(server, "*** %s: %s\n", _("Error"), text);
	}
	
	g_free(text);
}

/* say2_hello
 * Handle hello command
 */
static void
say2_hello(Say2Server *server, const EnamlBlock *hello)
{
	g_return_if_fail(server);

	/* Should only get a HELLO when in connected state */
	if (server->state != SAY2_SERVER_CONNECTED) {
		chat_interface_say(server, "*** %s\n", _("Unexpected hello"));
		return;
	}

	/* Now in announced mode */
	server->state = SAY2_SERVER_ANNOUNCED;

	/* If we can send a login block, do so, otherwise say that they need
	 * to /login or /guest
	 */
	if ((server->me->id && server->server_pass)
	    || (server->me->nick && server->me->email)) {
		say2_server_send_login(server);
	} else {
		chat_interface_say(server, "*** %s\n",
				   _("Please /login or /guest"));
	}
}

/* say2_server_info
 * Handle server-info command
 */
static void
say2_server_info(Say2Server *server, const EnamlBlock *server_info)
{
	const gchar *server_prog, *version;
	gchar       *text, *uptime;

	g_return_if_fail(server);

	/* Need a block of information */
	if (!server_info) {
		chat_interface_say(server, "*** server-info: %s\n",
				   _("Bogus command format received"));
		return;
	}

	/* Get the information out of the block */
	server_prog = enaml_get_string(server_info, "server", "?");
	version = enaml_get_string(server_info, "version", "?");
	text = utf8_decode(enaml_get_string(server_info, "text", NULL));
	uptime = whois_timespan_str(time(NULL)
			            - enaml_get_long(server_info, "uptime", 0));

	/* Display the info, with the optional text on a second line */
	chat_interface_say(server, "*** %s/%s up for %s\n",
			   server_prog, version, uptime);
	if (text && strlen(text))
		chat_interface_say(server, "*** %s\n", text);

	g_free(text);
	g_free(uptime);
}

/* say2_motd
 * Handle motd command
 */
static void
say2_motd(Say2Server *server, const EnamlBlock *motd)
{
	gchar *text;

	g_return_if_fail(server);

	/* If there's no text, just silently ignore it */
	if (!motd)
		return;

	/* Decode the text and output it */
	text = utf8_decode(enaml_get_string(motd, "text", NULL));
	if (text)
		chat_interface_say(server, "*** [MOTD] %s\n", text);

	g_free(text);
}

/* say2_welcome
 * Handle welcome command
 */
static void
say2_welcome(Say2Server *server, const EnamlBlock *welcome)
{
	const gchar   *id;
	gchar         *nick;
	ChatInterface *ci;

	g_return_if_fail(server);

	/* Need a block of information */
	if (!welcome) {
		chat_interface_say(server, "*** welcome: %s\n",
				   _("Bogus command format received"));
		chat_interface_update(server, CHAT_UPDATE_CLOSE, NULL);
		return;
	}

	/* Get the information out of the block */
	nick = utf8_decode(enaml_get_string(welcome, "nick", server->me->nick));
	id = enaml_get_string(welcome, "id", server->me->id);

	/* Copy the nickname and id */
	if (nick) {
		/* Literally just copy, its an allocated string anyway */
		server->me->nick = nick;
	}
	if (id) {
		g_free(server->me->id);
		server->me->id = g_strdup(id);
	}

	/* Welcome them to the server */
	chat_interface_say(server, "*** %s, %s !\n", _("Welcome"),
			   server->me->nick);
	if (strlen(server->me->id))
		chat_interface_say(server, "*** %s %s\n", _("Your ID here is"),
				   server->me->id);
	
	/* Active now, let them command away. */
	server->state = SAY2_SERVER_ACTIVE;

	/* Update our information */
	chat_interface_update(server, CHAT_UPDATE_USER, server->me);

	/* Join the default channel.
	 * FIXME should make an interface if this fails
	 */
	ci = chat_interface_get(server, CHAT_RECIPTYPE_CHAN, NULL);

	/* Ensure notify is on, and get list of clients (will inherently
	 * whois ourself)
	 */
	send_enaml_buildf(server->sock, "prefs", "notify:%s", "on");
	send_enaml_buildf(server->sock, "list-clients", "type:%s by-id",
			  "user");
	server->squelch_prefs++;
	server->squelch_members++;
	server->members_do_whois++;
}

/* say2_bad_nick
 * Handle bad-nick command
 */
static void
say2_bad_nick(Say2Server *server, const EnamlBlock *bad_nick)
{
	gchar *nick, *text;

	g_return_if_fail(server);

	/* No block, then just error at them */
	if (!bad_nick) {
		chat_interface_say(server, "*** %s\n",
				   _("Can't use nickname"));
		return;
	}

	/* Get the information out of the block */
	nick = utf8_decode(enaml_get_string(bad_nick, "nick",
					    server->me->nick));
	text = utf8_decode(enaml_get_string(bad_nick, "text",
				            _("(no reason given)")));

	/* Output the error, followed by what they should do */
	chat_interface_say(server, "*** %s (%s): %s\n", _("Can't use nickname"),
			   nick, text);
	chat_interface_say(server, "*** %s\n", _("Please /login or /guest"));

	g_free(nick);
	g_free(text);
}

/* say2_ping
 * Handle ping command
 */
static void
say2_ping(Say2Server *server, const EnamlBlock *ping)
{
	EnamlBlock *block;
	
	g_return_if_fail(server);

	/* I think there's more to be gained by echoing back the whole
	 * block than just replying with a blank one
	 */
	block = enaml_new();
	if (ping) {
		enaml_add_block(block, "ping",
				enaml_copy_block(ping), ENAML_FREE);
	} else {
		enaml_add_flag(block, "ping");
	}
	send_enaml_block(server->sock, block);
	enaml_free(block);
}

/* say2_signon
 * Handle the signon command
 */
static void
say2_signon(Say2Server *server, const EnamlBlock *signon)
{
	const gchar *id, *email;
	gchar       *nick;
	gchar        timestr[12];
	time_t       msgtime;

	g_return_if_fail(server);

	/* Need a block of information */
	if (!signon) {
		chat_interface_say(server, "*** signon: %s\n",
				   _("Bogus command format received"));
		return;
	}

	/* Get the information out of the block */
	id = enaml_get_string(signon, "id", "?");
	nick = utf8_decode(enaml_get_string(signon, "nick", "?"));
	email = enaml_get_string(signon, "email", "?");
	msgtime = (time_t)enaml_get_long(signon, "time", time(NULL));

	/* Register the new user */
	say2_server_userlist_update(server, SAY2_USERLIST_NEW,
				    id, (void *)nick);

	/* Convert the time into a string */
	strcpy(timestr, "");
	strftime(timestr, sizeof timestr, "[%H:%M:%S] ", localtime(&msgtime));

	/* Output */
	chat_interface_say(server, "%s*** %s: %s <%s>\n",
			   timestr, _("Signon"), nick, email);

	/* Send a whois to get more information about them */
	send_enaml_buildf(server->sock, "whois", "id:%s type:%s icon",
			  id, "user");
	server->squelch_whois++;

	g_free(nick);
}

/* say2_signoff
 * Handle the signoff command
 */
static void
say2_signoff(Say2Server *server, const EnamlBlock *signoff)
{
	const gchar *id;
	gchar       *nick;
	gchar        timestr[12];
	time_t       msgtime;

	g_return_if_fail(server);

	/* Need a block of information */
	if (!signoff) {
		chat_interface_say(server, "*** signoff: %s\n",
				   _("Bogus command format received"));
		return;
	}

	/* Get the information out of the block */
	id = enaml_get_string(signoff, "id", "?");
	nick = utf8_decode(enaml_get_string(signoff, "nick", "?"));
	msgtime = (time_t)enaml_get_long(signoff, "time", time(NULL));

	/* Remove the user from the lists */
	say2_server_userlist_update(server, SAY2_USERLIST_REMOVE, id, NULL);

	/* Convert the time into a string */
	strcpy(timestr, "");
	strftime(timestr, sizeof timestr, "[%H:%M:%S] ", localtime(&msgtime));

	/* Output */
	chat_interface_say(server, "%s*** %s: %s\n", timestr,
			   _("Signoff"), nick);

	g_free(nick);
}

/* say2_prefs
 * Handle the prefs command
 */
static void
say2_prefs(Say2Server *server, const EnamlBlock *prefs)
{
	const gchar *notify;
	gchar       *tagline;
	gint         password;
	
	g_return_if_fail(server);

	/* Need a block of information */
	if (!prefs) {
		chat_interface_say(server, "*** prefs: %s\n",
				   _("Bogus command format received"));
		return;
	}

	/* Get the information out of the block */
	notify = enaml_get_string(prefs, "notify", "?");
	tagline = utf8_decode(enaml_get_string(prefs, "tagline", NULL));
	password = (enaml_type(prefs, "password-changed") != ENAML_KEY_UNDEF
		    ? 1 : 0);

	/* Save the tagline */
	if (tagline) {
		g_free(server->me->tagline);
		server->me->tagline = g_strdup(tagline);
	}
	
	if (!server->squelch_prefs) {
		/* Display new preferences */
		chat_interface_say(server, "*** %s: %s = %s\n",
				   _("Preferences"), _("Notify"), notify);
		if (tagline) {
			chat_interface_say(server, "*** %s: %s = %s\n",
					   _("Preferences"), _("Tagline"),
					   tagline);
		}
		if (password)
			chat_interface_say(server, "*** %s: %s\n",
					   _("Preferences"),
					   _("Password Changed"));
	} else {
		/* Don't display anything */
		server->squelch_prefs--;
	}

	g_free(tagline);
}

/* say2_status
 * Handle the status command
 */
static void
say2_status(Say2Server *server, const EnamlBlock *status)
{
	const gchar *id, *mode;
	gchar       *nick, *text, *who;
	void        *info[2];

	g_return_if_fail(server);

	/* Need a block of information */
	if (!status) {
		chat_interface_say(server, "*** status: %s\n",
				   _("Bogus command format received"));
		return;
	}

	/* Get the information out of the block */
	id = enaml_get_string(status, "id", server->me->id);
	nick = utf8_decode(enaml_get_string(status, "nick", server->me->nick));
	mode = enaml_get_string(status, "mode", "active");
	text = utf8_decode(enaml_get_string(status, "text", NULL));

	/* If it was us that's changing status, adjust the message accordingly
	 * to make the english nice
	 */
	if (!g_strcasecmp(id, server->me->id)) {
		who = g_strdup(_("You are now"));
	} else {
		who = g_strdup_printf("%s %s", nick, _("is now"));
	}

	/* Update the information */
	info[0] = (void *)mode;
	info[1] = (void *)text;
	say2_server_userlist_update(server, SAY2_USERLIST_STATUS, id, info);

	/* Long winded to aid i18n */
	if (!g_strcasecmp(mode, "active")) {
		mode = _("here");
	} else if (!g_strcasecmp(mode, "away")) {
		mode = _("away");
	} else if (!g_strcasecmp(mode, "busy")) {
		mode = _("busy");
	} else if (!g_strcasecmp(mode, "idle")) {
		mode = _("idle");
	} else {
		mode = "?";
	}

	/* Output */
	if (text) {
		chat_interface_say(server, "*** %s %s: %s\n", who, mode, text);
	} else {
		chat_interface_say(server, "*** %s %s\n", who, mode);
	}
	
	g_free(who);
	g_free(nick);
	g_free(text);
}

/* say2_change_nick
 * Handle the change-nick command
 */
static void
say2_change_nick(Say2Server *server, const EnamlBlock *change_nick)
{
	const gchar *id;
	gchar       *old_nick, *nick;
	gchar       *who;

	g_return_if_fail(server);

	/* Need a block of information */
	if (!change_nick) {
		chat_interface_say(server, "*** change-nick: %s\n",
				   _("Bogus command format received"));
		return;
	}

	/* Get the information out of the block */
	id = enaml_get_string(change_nick, "id", server->me->id);
	old_nick = utf8_decode(enaml_get_string(change_nick, "old-nick",
						server->me->nick));
	nick = utf8_decode(enaml_get_string(change_nick, "nick", "?"));

	/* Update the information */
	say2_server_userlist_update(server, SAY2_USERLIST_NICK,
				    id, (void *)nick);

	/* If it was us that's changing status, adjust the message accordingly
	 * to make the english nice
	 */
	if (!g_strcasecmp(id, server->me->id)) {
		who = g_strdup(_("You are now known as"));
	} else {
		who = g_strdup_printf("%s %s", old_nick, _("is now known as"));
	}

	/* Output */
	chat_interface_say(server, "*** %s %s\n", who, nick);

	g_free(who);
	g_free(nick);
	g_free(old_nick);
}

/* say2_members
 * Handle the members command
 */
static void
say2_members(Say2Server *server, const EnamlBlock *members)
{
	const gchar *type;
	EnamlList   *list;
	GString     *buffer;
	gint         do_whois;

	g_return_if_fail(server);

	/* Need a block of information */
	if (!members) {
		chat_interface_say(server, "*** members: %s\n",
				   _("Bogus command format received"));
		return;
	}

	/* Get the information out of the block */
	type = enaml_get_string(members, "type", "user");
	list = enaml_get_list(members, "list");

	/* If this is a user members list, and members_do_whois is set, this
	 * is most likely the reply to the list-clients done when we joined.
	 * In that case whois each user that we find
	 */
	if (server->members_do_whois && !g_strcasecmp(type, "user")) {
		do_whois = 1;
		server->members_do_whois--;
	} else {
		do_whois = 0;
	}

	/* Allocate space for the output buffer */
	buffer = g_string_new("");
	if (list) {
		gint i, count;

		/* Iterate the list of members */
		count = enaml_list_length(list);
		for (i = 0; i < count; i++) {
			const gchar *member;
			gchar       *member2;

			/* Get the member */
			member = enaml_list_get_string(list, i, "?");

			/* Add the member name to the output buffer.  We'll
			 * assume for this case that the member is a nickname
			 * (because that's all we allow them to get from this
			 * command) so it's UTF-8 */
			member2 = utf8_decode(member);
			g_string_sprintfa(buffer, " %s", member2);
			g_free(member2);

			/* Whois each member, we know that this means that
			 * members is a list of ids rather than nicks, so
			 * we'll take the not-utf-8 version just for safety
			 */
			if (do_whois) {
				/* It's safe to call new, if they exist
				 * this won't do anything
				 */
				say2_server_userlist_update(server,
							    SAY2_USERLIST_NEW,
							    member, NULL);

				/* Send the WHOIS */
				send_enaml_buildf(server->sock, "whois",
						  "id:%s type:%s icon",
						  member, "user");
				server->squelch_whois++;
			}
		}
	}

	/* Display the members reply */
	if (!server->squelch_members) {
		chat_interface_say(server, "*** %s:%s\n",
				   _("Members"), buffer->str);
	} else {
		server->squelch_members--;
	}

	g_string_free(buffer, TRUE);
}

/* say2_begin_response
 * Handle the begin-response command
 */
static void
say2_begin_response(Say2Server *server, const EnamlBlock *begin_response)
{
	const gchar *response;
	glong        count, total;
	
	g_return_if_fail(server);

	/* Need a block of information */
	if (!begin_response) {
		chat_interface_say(server, "*** begin-response: %s\n",
				   _("Bogus command format received"));
		return;
	}

	/* Get the information out of the block */
	response = enaml_get_string(begin_response, "response", "?");
	count = enaml_get_long(begin_response, "count", 0);
	total = enaml_get_long(begin_response, "total", 0);

	if (!g_strcasecmp(response, "client-info")) {
		/* Whom reply */

		/* Ditch any current one */
		whois_info_free_list(server->whom_list);

		/* Fill in info */
		server->in_whom = TRUE;
		server->whom_list = NULL;
		server->whom_count = count;
		server->whom_total = total;

	} else if (!g_strcasecmp(response, "say")) {
		/* History reply */
		server->in_history = TRUE;
		chat_interface_say(server, "*** %s:\n", _("History begins"));

	} else {
		/* Sodomy non sapiens */
		chat_interface_say(server, "*** %s (%s)\n",
				   _("Unknown response block"), response);
	}
}

/* say2_end_response
 * Handle the end-response command
 */
static void
say2_end_response(Say2Server *server, const EnamlBlock *end_response)
{
	const gchar *response;
	
	g_return_if_fail(server);

	/* Need a block of information */
	if (!end_response) {
		chat_interface_say(server, "*** end-response: %s\n",
				   _("Bogus command format received"));
		return;
	}

	/* Get the information out of the block */
	response = enaml_get_string(end_response, "response", "?");

	if (!g_strcasecmp(response, "client-info")) {
		/* End of Whom */
		server->whom_list = g_slist_reverse(server->whom_list);
		whois_info_dump_list(server, server->whom_list,
				     server->whom_count, server->whom_total);

		/* Clean up */
		whois_info_free_list(server->whom_list);
		server->whom_list = NULL;
		server->in_whom = FALSE;

	} else if (!g_strcasecmp(response, "say")) {
		/* End of History */
		server->in_history = FALSE;
		chat_interface_say(server, "*** %s.\n", _("End of History"));

	} else {
		/* Sodomy non sapiens */
		chat_interface_say(server, "*** %s (%s)\n",
				   _("Unknown response block end"), response);
	}
}

/* say2_client_info
 * Handle the client-info command
 */
static void
say2_client_info(Say2Server *server, const EnamlBlock *client_info)
{
	WhoisInfo *wi;

	g_return_if_fail(server);

	/* Need a block of information */
	if (!client_info) {
		chat_interface_say(server, "*** client-info: %s\n",
				   _("Bogus command format received"));
		return;
	}

	/* Reap the information */
	wi = whois_info_new_from_block(client_info);

	/* If it's a user, we want to update their information with the new
	 * stuff
	 */
	if (!g_strcasecmp(wi->type, "user")) {
		/* Do the update */
		say2_server_userlist_update(server, SAY2_USERLIST_WHOIS,
					    wi->id, (void *)wi);

		/* User whois's may need to be squelched */
		if (server->squelch_whois) {
			server->squelch_whois--;
			goto out;
		}
	}

	if (server->in_whom) {
		/* Actually part of a WHOM */
		server->whom_list = g_slist_prepend(server->whom_list, wi);
		return;
	} else {
		/* Display it */
		whois_info_dump(server, wi);
	}

out:
	whois_info_free(wi);
}

/* say2_not_present
 * Handle the not-present command
 */
static void
say2_not_present(Say2Server *server, const EnamlBlock *not_present)
{
	const gchar *id, *type;
	gchar       *nick;

	g_return_if_fail(server);

	/* If server doesn't say who, it doesn't particularly matter */
	if (!not_present) {
		chat_interface_say(server, "*** %s.\n", _("Not logged in"));
		return;
	}

	/* Get the information out of the block */
	id = enaml_get_string(not_present, "id", "?");
	nick = utf8_decode(enaml_get_string(not_present, "nick", NULL));
	type = enaml_get_string(not_present, "type", "user");

	/* Output */
	chat_interface_say(server, "*** %s %s.\n", (nick ? nick : id),
			   _("is not logged in"));

	g_free(nick);
}

/* say2_no_such_client
 * Handle the no-such-client command
 */
static void
say2_no_such_client(Say2Server *server, const EnamlBlock *no_such_client)
{
	const gchar *id, *type;

	g_return_if_fail(server);

	/* If server doesn't say who, it doesn't particularly matter */
	if (!no_such_client) {
		chat_interface_say(server, "*** %s.\n", _("No such client"));
		return;
	}

	/* Get the information out of the block */
	id = enaml_get_string(no_such_client, "id", "?");
	type = enaml_get_string(no_such_client, "type", "user");

	/* Output */
	if (!g_strcasecmp(type, "user")) {
		chat_interface_say(server, "*** %s: %s\n", _("No such user"),
				   id);
	} else if (!g_strcasecmp(type, "channel")) {
		chat_interface_say(server, "*** %s: %s\n", _("No such channel"),
				   id);
	} else {
		chat_interface_say(server, "*** %s: %s\n", _("No such client"),
				   id);
	}
}

#if 0
/* say2_client_attributes
 * Handle the client-attributes command
 */
static void
say2_client_attributes(Say2Server *server, const EnamlBlock *client_attributes)
{
	/* FIXME */
}

/* say2_set_attributes
 * Handle the set-attributes command
 */
static void
say2_set_attributes(Say2Server *server, const EnamlBlock *set_attributes)
{
	/* FIXME */
}
#endif

/* say2_say
 * Handle the say command
 */
static void
say2_say(Say2Server *server, const EnamlBlock *say)
{
	const gchar *id;
	gchar       *nick, *text, *act;
	gchar        timestr[40];
	time_t       msgtime;

	g_return_if_fail(server);

	/* Need a block of information */
	if (!say) {
		chat_interface_say(server, "*** say: %s\n",
				   _("Bogus command format received"));
		return;
	}

	/* Get the information out of the block */
	id = enaml_get_string(say, "id", "?");
	nick = utf8_decode(enaml_get_string(say, "nick", "?"));
	text = utf8_decode(enaml_get_string(say, "text", NULL));
	act = utf8_decode(enaml_get_string(say, "act", NULL));
	msgtime = (time_t)enaml_get_long(say, "time", time(NULL));

	/* Convert the time into a string */
	strcpy(timestr, "");
	if (server->in_history) {
		time_t now, diff;

		now = time(NULL);
		diff = now - msgtime;

		if (diff < 7200L) {
			/* Within 2 hours [hh:mm:ss] */
			strftime(timestr, sizeof timestr, "[%H:%M:%S] ",
				 localtime(&msgtime));
		} else if (diff < 82800L) {
			/* Within 23 hours [hh:mm] */
			strftime(timestr, sizeof timestr, "[%H:%M] ",
				 localtime(&msgtime));
		} else if (diff < 518400L) {
			/* Within 6 days [day hh:mm] */
			strftime(timestr, sizeof timestr, "[%a %H:%M] ",
				 localtime(&msgtime));
		} else if (diff < 25920000L) {
			/* Within 300 days [d mon] */
			strftime(timestr, sizeof timestr, "[%d %b] ",
				 localtime(&msgtime));
		} else {
			/* Otherwise [d mon yyyy] */
			strftime(timestr, sizeof timestr, "[%d %b %Y] ",
				 localtime(&msgtime));
		}
	} else {
		strftime(timestr, sizeof timestr, "[%H:%M:%S] ",
			 localtime(&msgtime));
	}

	/*
	 * I think commands should be able to have both act's and text's
	 * and any other type of message in them (just like IRC can have both
	 * text and multiple CTCP in the same message) so don't else this and
	 * treat as seperate ifs.
	 */
	if (act)
		chat_interface_say(server, "%s* %s %s\n", timestr, nick, act);
	if (text)
		chat_interface_say(server, "%s<%s> %s\n", timestr, nick, text);

	g_free(nick);
	g_free(text);
	g_free(act);
}

/* say2_message
 * Handle the message command
 */
static void
say2_message(Say2Server *server, const EnamlBlock *say)
{
	const gchar *id, *to_id;
	gchar       *nick, *to_nick, *text, *act;
	gchar        timestr[12];
	time_t       msgtime;

	g_return_if_fail(server);

	/* Need a block of information */
	if (!say) {
		chat_interface_say(server, "*** message: %s\n",
				   _("Bogus command format received"));
		return;
	}

	/* Get the information out of the block */
	id = enaml_get_string(say, "id", "?");
	to_id = enaml_get_string(say, "to-id", "?");
	nick = utf8_decode(enaml_get_string(say, "nick", "?"));
	to_nick = utf8_decode(enaml_get_string(say, "to-nick", "?"));
	text = utf8_decode(enaml_get_string(say, "text", NULL));
	act = utf8_decode(enaml_get_string(say, "act", NULL));
	msgtime = (time_t)enaml_get_long(say, "time", time(NULL));

	/* Convert the time into a string */
	strcpy(timestr, "");
	strftime(timestr, sizeof timestr, "[%H:%M:%S] ", localtime(&msgtime));

	/* Whenever we send or receive a message, the server sends us a 
	 * command to indicate that.  If we send a message to ourselves, the
	 * server will only give us one command, so do these next two seperately
	 * to show it both going out and coming back (which we want to see).
	 *
	 * Also I think commands should be able to have both act's and text's
	 * and any other type of message in them (just like IRC can have both
	 * text and multiple CTCP in the same message) so don't else that if
	 * either
	 */
	if (!g_strcasecmp(id, server->me->id)) {
		/* From me */
		if (act) {
			chat_interface_say(server, "%s-> [%s] * %s %s\n",
					   timestr, to_nick, nick, act);
		}
		if (text) {
			chat_interface_say(server, "%s-> [%s] %s\n",
					   timestr, to_nick, text);
		}
	}
	if (!g_strcasecmp(to_id, server->me->id)) {
		/* To me */
		if (act) {
			chat_interface_say(server, "%s* [%s] %s\n",
					   timestr, nick, act);
		}
		if (text) {
			chat_interface_say(server, "%s[%s] %s\n",
					   timestr, nick, text);
		}
	}

	g_free(nick);
	g_free(to_nick);
	g_free(text);
	g_free(act);
}

/* say2_wall
 * Handle the wall command
 */
static void
say2_wall(Say2Server *server, const EnamlBlock *wall)
{
	const gchar *id;
	gchar       *nick, *text;
	gchar        timestr[12];
	time_t       msgtime;

	g_return_if_fail(server);

	/* Need a block of information */
	if (!wall) {
		chat_interface_say(server, "*** wall: %s\n",
				   _("Bogus command format received"));
		return;
	}

	/* Get the information out of the block */
	id = enaml_get_string(wall, "id", "?");
	nick = utf8_decode(enaml_get_string(wall, "nick", "?"));
	text = utf8_decode(enaml_get_string(wall, "text", NULL));
	msgtime = (time_t)enaml_get_long(wall, "time", time(NULL));

	/* Convert the time into a string */
	strcpy(timestr, "");
	strftime(timestr, sizeof timestr, "[%H:%M:%S] ", localtime(&msgtime));

	/* Output */
	chat_interface_say(server, "%s*** %s %s: %s\n", timestr, _("WALL from"),
			   nick, text);

	g_free(nick);
	g_free(text);
}


/* say2_dispatch_go
 * Dispatch a command from the say2 command_table
 */
gboolean
say2_dispatch_go(const gchar *command, Say2Server *server,
		 const EnamlBlock *sub)
{
	static void *dispatch_table = NULL;

	if (!dispatch_table)
		dispatch_table = dispatch_create(command_table);
	
	return dispatch_go(dispatch_table, command, server, (void *)sub, NULL);
}
