/* meko
 * Copyright (C) 2001 Scott James Remnant <scott@netsplit.com>.
 * All Rights Reserved.
 *
 * say2_server.c
 *  - Handles the Say2Client structure which contains information about
 *    our server connection
 *  - Communication with a say2 server
 * -- 
 * $Id: say2_server.c,v 1.18 2001/06/17 16:48:39 scott Exp $
 * 
 * This file is distributed according to the GNU General Public
 * License.  For full details, read the top of 'main.c' or the
 * file called COPYING that was distributed with this code.
 */

#include <meko.h>

#include <netdb.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <glib.h>
#include <gnome.h>
#include <say2/say2.h>

#include "say2_server.h"
#include "say2_client.h"
#include "say2_commands.h"
#include "chat_interface.h"
#include "unicode.h"

/* Forward prototypes for socket callback functions */
static void say2_server_read(Socket *);
static void say2_server_ready(Socket *, gint);
static void say2_server_eof(Socket *);

/* Forward prototypes for the hash entry free function */
static gboolean say2_server_free_users_entry(gchar *, Say2Client *,
					     Say2Server *);

/* say2_server_new
 * Create a Say2Server structure and connected to the given hostname and
 * optional port.  Should be followed by a call to set_login or set_guest
 * to ensure we don't login as something stupid.
 */
Say2Server *
say2_server_new(const gchar *hostname, const gchar *port)
{
	Say2Server     *server;
	struct servent *servent;
	gint            portnum;

	g_return_val_if_fail(hostname && strlen(hostname), NULL);

	/* If port isn't specified, just point at DEFAULT_PORT */
	if (!port || !strlen(port))
		port = DEFAULT_SERVER_PORT;

	/* Get the numerical value of the port */
	servent = getservbyname(port, "tcp");
	if (servent) {
		portnum = servent->s_port;
	} else {
		portnum = atoi(port);
	}

	/* No luck, abort */
	if (!portnum) {
		chat_interface_say(NULL, "*** %s: %s\n",
				   _("Invalid port"), port);
		return NULL;
	}
	
	/* Allocate space for everything */
	server = g_new0(Say2Server, 1);
	server->me = say2_client_new();
	server->users = g_hash_table_new(g_str_hash, g_str_equal);

	/* Fill name and port from our paremters */
	server->server_name = g_strdup(hostname);
	server->server_port = portnum;

	/* Fill our id with something silly for now */
	say2_client_ref(server->me);
	server->me->id = g_strdup("");

	/* Time to connect, all output is sent to a NULL server.  We can
	 * fairly safely assume that one of those will be assigned to us
	 * as soon as we've finished
	 */
	chat_interface_say(NULL, "*** %s %s %s %d...\n",
			   _("Connecting to"), server->server_name,
			   _("port"), server->server_port);

	/* Create the socket */
	server->sock = socket_new(NULL, 0);
	if (!server->sock) {
		chat_interface_say(NULL, "*** %s. :(\n",
				   _("Failed to make socket"));
		goto error;
	}
	socket_set_data(server->sock, server);

	/* Connect to it */
	if (!socket_connect(server->sock, server->server_name,
			    server->server_port, &say2_server_ready)) {
		chat_interface_say(NULL, "*** %s. :(\n",
				   _("Failed to connect"));
		socket_close(server->sock);
		goto error;
	}

	/* Our work here is done */
	server->state = SAY2_SERVER_CONNECTING;
	return server;

error:
	say2_client_unref(server->me);
	g_hash_table_destroy(server->users);
	g_free(server->server_name);
	g_free(server);

	return NULL;
}

/* say2_server_set_login
 * Set the server up to login to the server with an id and password
 */
void
say2_server_set_login(Say2Server *server, const gchar *id, const gchar *pass)
{
	g_return_if_fail(server);
	g_return_if_fail(id && strlen(id));
	g_return_if_fail(pass && strlen(pass));

	/* Free the existing information */
	g_free(server->server_pass);
	g_free(server->me->id);
	g_free(server->me->nick);

	/* Store new information */
	server->me->id = g_strdup(id);
	server->server_pass = g_strdup(pass);
	server->me->nick = 0;
	
	/* If the server is in announced state, we've most likely been refused
	 * a login and are trying again, so resend the command
	 */
	if (server->state == SAY2_SERVER_ANNOUNCED)
		say2_server_send_login(server);
}

/* say2_server_set_guest
 * Set the server up to login to the server with as a guest with a nickname
 * and e-mail address
 */
void
say2_server_set_guest(Say2Server *server, const gchar *nick, const gchar *email,
		      const gchar *tagline)
{
	g_return_if_fail(server);
	g_return_if_fail(nick && strlen(nick));
	g_return_if_fail(email && strlen(email));

	/* Free the existing information */
	g_free(server->server_pass);
	g_free(server->me->id);
	g_free(server->me->nick);
	g_free(server->me->email);
	g_free(server->me->tagline);

	/* Store new information */
	server->server_pass = 0;
	server->me->id = g_strdup("");
	server->me->nick = g_strdup(nick);
	server->me->email = g_strdup(email);
	server->me->tagline = g_strdup(tagline);
	
	/* If the server is in announced state, we've most likely been refused
	 * a login and are trying again, so resend the command
	 */
	if (server->state == SAY2_SERVER_ANNOUNCED)
		say2_server_send_login(server);
}

/* say2_server_send_login
 * Send a correctly and fully structured login command to the server based on
 * the current information
 */
void
say2_server_send_login(Say2Server *server)
{
	EnamlBlock *login, *block;

	g_return_if_fail(server);
	g_assert(server->state == SAY2_SERVER_ANNOUNCED);
	g_assert((server->me->id && server->server_pass)
		 || (server->me->nick && server->me->email));

	/* Construct the login block */
	login = enaml_new();
	enaml_add_string(login, "client", APPNAME " " FULL_VERSION, ENAML_COPY);

	/* Add either id/pass or nick/email/tagline */
	if (server->server_pass) {
		enaml_add_string(login, "nick", server->me->id, ENAML_COPY);
		enaml_add_string(login, "password", server->server_pass,
				 ENAML_COPY);
	} else {
		enaml_add_string(login, "nick", server->me->nick, ENAML_COPY);
		enaml_add_string(login, "email", server->me->email, ENAML_COPY);

		if (server->me->tagline) {
			gchar *tagline;

			tagline = utf8_encode(server->me->tagline);
			enaml_add_string(login, "tagline", tagline, ENAML_COPY);
			g_free(tagline);
		}
	}

	/* Construct the block to send it all */
	block = enaml_new();
	enaml_add_block(block, "login", login, ENAML_FREE);
	send_enaml_block(server->sock, block);
	enaml_free(block);
}

/* say2_server_read
 * Called when there's data on the socket.  Reads it, parses it and dispatches
 * it to the right function to handle the command.
 */
static void
say2_server_read(Socket *sock)
{
	Say2Server *server;

	/* Get the server structure off the socket */
	server = socket_get_data(sock);
	g_assert(server);
	g_assert(server->sock == sock);

	/* Loop through any lines available */
	for (;;) {
		EnamlBlock  *block;
		EnamlIter    iter;
		gchar       *line;
		const gchar *key;

		/* Get a line from the socket. */
		line = socket_getline(sock);
		if (!line)
			break;

		/* Parse the enaml in the line */
		block = enaml_parse(line);
		if (!block) {
			chat_interface_say(server, "*** %s: %s\n",
					   _("Bad Enaml"), line);
			goto out;
		}

		/* A line can contain lots of different enaml keys and blocks.
		 * Iterate them so we don't miss one
		 */
		for (key = enaml_iter_first(block, &iter); key;
		     key = enaml_iter_next(&iter)) {
			const EnamlBlock *sub;

			/* Get the block and try to dispatch it */
			sub = enaml_get_block(block, key);
			if (!say2_dispatch_go(key, server, sub)) {
				if (sub) {
					gchar *s;

					s = enaml_compose(block, FALSE);
					chat_interface_say(server,
							   "*** %s %s: %s\n",
							   _("Unknown Command"),
							   key, s);
					g_free(s);
				} else {
					chat_interface_say(server,
							   "*** %s: %s\n",
							   _("Unknown Command"),
							   key);
				}

				goto out;
			}
		}

out:
		/* Free up what we did */
		enaml_free(block);
		g_free(line);
	}
}

/* say2_server_ready
 * Called when a connection to the server finishes (either successfully or
 * not)
 */
static void
say2_server_ready(Socket *sock, gint success)
{
	Say2Server *server;

	/* Get the server structure off the socket */
	server = socket_get_data(sock);
	g_assert(server);
	g_assert(server->sock == sock);

	/* Connection failed */
	if (!success) {
		chat_interface_say(server, "*** %s. :(\n",
				   _("Failed to connect"));

		/* Close the socket */
		socket_close(server->sock);
		server->sock = 0;

		/* Send a CLOSE message to all linked chat interfaces.  This
		 * should reduce our reference count to 0 and free us (make
		 * sure that it does though).
		 */
		say2_server_ref(server);
		chat_interface_update(server, CHAT_UPDATE_CLOSE, NULL);
		g_assert(server->ref_count == 1);
		say2_server_unref(server);

		return;
	}

	chat_interface_say(server, "*** %s!\n", _("Connected"));

	server->state = SAY2_SERVER_CONNECTED;

	socket_set_eof_fn(sock, &say2_server_eof);
	socket_set_read_fn(sock, &say2_server_read);
}

/* say2_server_eof
 * Called when the server drops our connection.
 */
static void
say2_server_eof(Socket *sock)
{
	Say2Server *server;

	/* Get the server structure off the socket */
	server = socket_get_data(sock);
	g_assert(server);
	g_assert(server->sock == sock);
	
	chat_interface_say(server, "*** %s.\n", _("Server dropped link"));
	
	/* Mark the socket as closed */
	server->sock = 0;

	/* Send a CLOSE message to all linked chat interfaces.  This should
	 * reduce our reference count to 0 and free us (make sure that it
	 * does though).
	 */
	say2_server_ref(server);
	chat_interface_update(server, CHAT_UPDATE_CLOSE, NULL);
	g_assert(server->ref_count == 1);
	say2_server_unref(server);
}

/* say2_server_userlist_update
 * Add, update or remove the user list entry with the given id, and optional
 * second parameter for updates
 */
Say2Client *
say2_server_userlist_update(Say2Server *server, Say2ServerULUpdate update_type,
			    const gchar *id, void *extra)
{
	gchar      *hashkey, *key;
	Say2Client *client;
	gboolean    found;

	g_return_val_if_fail(server, NULL);
	g_return_val_if_fail(id && strlen(id), NULL);

	/* The hash key is a lowercase version of the id */
	g_strdown(hashkey = g_strdup(id));
	
	/* Nearly all of the options require it, so we'll look up the
	 * real key and data
	 */
	found = g_hash_table_lookup_extended(server->users, hashkey,
					     (gpointer *)&key,
					     (gpointer *)&client);

	/* If it wasn't in the users list, it still might be us, so at least
	 * keep a hold of that client
	 */
	if (!found) {
		if (!g_strcasecmp(id, server->me->id)) {
			client = server->me;
		} else {
			client = NULL;
		}
	}

	/* Update type switch */
	switch (update_type) {
	case SAY2_USERLIST_NEW:
		/* Indicates that the user should exist in the user list, and
		 * will be created if it doesn't
		 */
		if (!found) {
			/* If it wasn't us (client will be set to us if it
			 * was) then make a new client
			 */
			if (!client) {
				client = say2_client_new();
				client->id = g_strdup(id);
			}

			/* Optional argument is a nickname */
			if (extra) {
				const gchar *nick;

				nick = (const gchar *)extra;
				g_free(client->nick);
				client->nick = g_strdup(nick);
			}

			/* Add it to the users list and notify any chat
			 * interfaces about it */
			g_hash_table_insert(server->users, hashkey, client);
			chat_interface_update(server, CHAT_UPDATE_SIGNON,
					      client);

			/* Reference it, and make sure we don't free the
			 * hashkey in a bit (it's been copied to the table)
			 */
			say2_client_ref(client);
			hashkey = 0;
		}

		break;

	case SAY2_USERLIST_REMOVE:
		/* Indicates that the user should not exist in the user list,
		 * and will be removed if it does.  This won't try to remove
		 * us from the userlist unless we were in it.
		 */
		if (found) {
			g_assert(key);
			g_assert(client);

			/* Notify that they've signed off and free it */
			chat_interface_update(server, CHAT_UPDATE_SIGNOFF,
					      client);
			say2_server_free_users_entry(key, client, server);
			client = NULL;
		}

		break;
	
	case SAY2_USERLIST_STATUS:
		/* Indicates that the status of the client needs to be changed.
		 * This can be used to update server->me too, so doesn't rely
		 * on the entry being in the user list.  The extra parameter
		 * is required and must be a gchar ** to an array of two
		 * gchar's, the status mode and status message (which can be
		 * NULL
		 */
		if (client && extra) {
			const gchar **status;

			status = (const gchar **)extra;

			/* First parameter is the texual status mode, convert
			 * to an enum
			 */
			if (!g_strcasecmp(status[0], "active")) {
				client->status = SAY2_STATUS_ACTIVE;
			} else if (!g_strcasecmp(status[0], "busy")) {
				client->status = SAY2_STATUS_BUSY;
			} else if (!g_strcasecmp(status[0], "away")) {
				client->status = SAY2_STATUS_AWAY;
			} else {
				client->status = SAY2_STATUS_OTHER;
			}

			/* Second parameter is NULL or a utf8 decoded version
			 * of the status text
			 */
			g_free(client->status_msg);
			client->status_msg = g_strdup(status[1]);

			/* Notify that it's been updated */
			chat_interface_update(server, CHAT_UPDATE_USER, client);
		} else {
			g_assert(client);
			g_assert(extra);
		}

		break;
	
	case SAY2_USERLIST_NICK:
		/* Indicates that the nickname of the client needs to be
		 * changed.  This can be used to update server->me too, so
		 * doesn't rely on the entry being in the user list.  The
		 * extra parameter is required and must be a gchar * to the
		 * new nickname.
		 */
		if (client && extra) {
			const gchar *nick;

			/* Copy from nickname */
			nick = (const gchar *)extra;
			g_free(client->nick);
			client->nick = g_strdup(nick);

			/* Notify that it's been updated */
			chat_interface_update(server, CHAT_UPDATE_USER, client);
		} else {
			g_assert(client);
			g_assert(extra);
		}
			
		break;
	
	case SAY2_USERLIST_WHOIS:
		/* Indicates that a whois has been done on the client and
		 * we've got lots of info that may need to be changed.  This
		 * can be used to update server->me too, so doesn't rely on
		 * the entry being in the user list.  The extra parameter is
		 * required and must be a WhoisInfo * to the whois reply.
		 */
		if (client && extra) {
			const WhoisInfo *wi;

			/* Copy from WhoisInfo structure */
			wi = (const WhoisInfo *)extra;
			say2_client_copy_from_whois(client, wi);

			/* Notify that it's been updated */
			chat_interface_update(server, CHAT_UPDATE_USER, client);
		} else {
			g_assert(client);
			g_assert(extra);
		}

		break;
		
	default:
		/* NOTREACHED! */
		g_assert_not_reached();
	}

	/* Free up */
	g_free(hashkey);

	/* Return the client we've been playing with */
	return client;
}

/* say2_server_free_users_entry
 * Called to free up a hash table entry, because we need to actually free
 * both the key and unreference the client.
 */
static gboolean
say2_server_free_users_entry(gchar *id, Say2Client *client, Say2Server *server)
{
	g_free(id);
	say2_client_unref(client);
	return TRUE;
}

/* say2_server_ref
 * Increment the reference count of a Say2Server structure
 */
void
say2_server_ref(Say2Server *server)
{
	if (!server)
		return;

	server->ref_count++;
}

/* say2_server_unref
 * Decrement the reference count of a Say2Client structure, freeing it if
 * the reference count reaches 0
 */
void
say2_server_unref(Say2Server *server)
{
	if (!server)
		return;

	g_assert(server->ref_count);

	server->ref_count--;

	/* Reference count reached 0, free it */
	if (!server->ref_count) {
		/* Close the socket if its still open */
		if (server->sock) {
			/* Doing a callback right now would be bad, m'kay */
			socket_set_read_fn(server->sock, NULL);
			socket_set_eof_fn(server->sock, NULL);

			/* Close it, eof_fn() will be triggered, but that's
			 * ok, the CLOSE update won't unref us anymore so
			 * we won't end up getting recursive */
			socket_close(server->sock);
		}

		/* Free the key and unref each of our users entries */
		g_hash_table_foreach_remove
			(server->users, (GHRFunc)say2_server_free_users_entry,
			 (gpointer)server);
		g_hash_table_destroy(server->users);

		/* Free the whom list if we were doing that */
		whois_info_free_list(server->whom_list);

		/* Free other string information */
		g_free(server->server_name);
		g_free(server->server_pass);

		/* Dereference the client pointer */
		say2_client_unref(server->me);

		/* Free ourselves */
		g_free(server);
	}
}
