/* meko
 * Copyright (C) 2001 Scott James Remnant <scott@netsplit.com>.
 * All Rights Reserved.
 *
 * user_commands.c
 *  - Handles commands entered by the user
 * -- 
 * $Id: user_commands.c,v 1.9 2001/06/17 16:36:02 scott Exp $
 * 
 * This file is distributed according to the GNU General Public
 * License.  For full details, read the top of 'main.c' or the
 * file called COPYING that was distributed with this code.
 */

#include <meko.h>

#include <string.h>

#include <glib.h>
#include <gnome.h>
#include <gtk/gtk.h>
#include <say2/say2.h>

#include "user_commands.h"
#include "chat_interface.h"
#include "connect_dialog.h"
#include "say2_types.h"
#include "say2_server.h"
#include "unicode.h"

/* Forward prototype for command functions */
static void user_motd(ChatInterface *, const gchar *);
static void user_prefs(ChatInterface *, const gchar *);
static void user_password(ChatInterface *, const gchar *);
static void user_status(ChatInterface *, const gchar *);
static void user_away(ChatInterface *, const gchar *);
static void user_busy(ChatInterface *, const gchar *);
static void user_here(ChatInterface *, const gchar *);
static void user_nick(ChatInterface *, const gchar *);
static void user_whom(ChatInterface *, const gchar *);
static void user_who(ChatInterface *, const gchar *);
static void user_whois(ChatInterface *, const gchar *);
static void user_seen(ChatInterface *, const gchar *);
static void user_channel(ChatInterface *, const gchar *);
static void user_say(ChatInterface *, const gchar *);
static void user_me(ChatInterface *, const gchar *);
static void user_msg(ChatInterface *, const gchar *);
static void user_act(ChatInterface *, const gchar *);
static void user_wall(ChatInterface *, const gchar *);
static void user_history(ChatInterface *, const gchar *);
static void user_boot(ChatInterface *, const gchar *);
static void user_clear(ChatInterface *, const gchar *);
static void user_connect(ChatInterface *, const gchar *);
static void user_disconnect(ChatInterface *, const gchar *);
static void user_login(ChatInterface *, const gchar *);
static void user_guest(ChatInterface *, const gchar *);
static void user_close(ChatInterface *, const gchar *);
static void user_quit(ChatInterface *, const gchar *);

/* Command dispatch table */
static DispatchItem command_table[] = {
	{ "motd", &user_motd },
	{ "prefs", &user_prefs },
	{ "pref", &user_prefs },
	{ "password", &user_password },
	{ "pass", &user_password },
	{ "status", &user_status },
	{ "away", &user_away },
	{ "busy", &user_busy },
	{ "here", &user_here },
	{ "unaway", &user_here },
	{ "unbusy", &user_here },
	{ "nick", &user_nick },
	{ "whom", &user_whom },
	{ "who", &user_who },
	{ "members", &user_who },
	{ "whois", &user_whois },
	{ "seen", &user_seen },
	{ "whowas", &user_seen },
	{ "channel", &user_channel },
	{ "chan", &user_channel },
	{ "say", &user_say },
	{ "me", &user_me },
	{ "msg", &user_msg },
	{ "m", &user_msg },
	{ "action", &user_act },
	{ "act", &user_act },
	{ "wall", &user_wall },
	{ "history", &user_history },
	{ "hist", &user_history },
	{ "boot", &user_boot },
	{ "clear", &user_clear },
	{ "connect", &user_connect },
	{ "server", &user_connect },
	{ "disconnect", &user_disconnect },
	{ "logout", &user_disconnect },
	{ "discon", &user_disconnect },
	{ "close", &user_close },
	{ "quit", &user_quit },
	{ "exit", &user_quit },
#if 0
	{ "raw", &user_raw },
#endif
	{ NULL, NULL }
};

/* Command dispatch table for unconnected interfaces */
static DispatchItem uc_command_table[] = {
	{ "clear", &user_clear },
	{ "connect", &user_connect },
	{ "server", &user_connect },
	{ "disconnect", &user_disconnect },
	{ "logout", &user_disconnect },
	{ "discon", &user_disconnect },
	{ "login", &user_login },
	{ "guest", &user_guest },
	{ "close", &user_close },
	{ "quit", &user_quit },
	{ "exit", &user_quit },
	{ NULL, NULL }
};

/* Command prefix */
static const gchar command_prefix[] = "/";


/* user_motd
 * Handle the /motd command
 */
static void
user_motd(ChatInterface *ci, const gchar *text)
{
	g_return_if_fail(ci);

	/* Simply send a motd request to the server */
	send_enaml_buildf(ci->server->sock, "motd", "");
}

/* user_prefs
 * Handle the /prefs command
 */
static void
user_prefs(ChatInterface *ci, const gchar *text)
{
	gchar **arg;

	g_return_if_fail(ci);

	/* If we don't get the two required arguments, give them a quick
	 * Usage: message and then ask the server for our current prefs
	 */
	arg = g_strsplit(text, " ", 1);
	if (!arg[0] || !strlen(arg[0]) || !arg[1] || !strlen(arg[1])) {
		g_strfreev(arg);
		chat_interface_sayto(ci, "*** %s: %spref %s\n",
				     _("Usage"), command_prefix,
				     _("[ <pref name> <new value> ]"));
		send_enaml_buildf(ci->server->sock, "prefs", "");
		return;
	}

	/* Can only change the tagline for now.  Password is changed using
	 * /pass, notify and icon are used by meko
	 */
	if (!g_strcasecmp(arg[0], "tagline")) {
		gchar *text2;

		text2 = utf8_encode(arg[1] ? arg[1] : "");
		send_enaml_buildf(ci->server->sock, "prefs", "tagline:%s",
				  text2);
		g_free(text2);

	} else {
		chat_interface_sayto(ci, "*** %s: %s\n",
				     _("Available preferences"), "tagline");
	}

	g_strfreev(arg);
}

/* user_password
 * Handle the /password command
 */
static void
user_password(ChatInterface *ci, const gchar *text)
{
	g_return_if_fail(ci);

	/* A new password must be supplied */
	if (!strlen(text)) {
		chat_interface_sayto(ci, "*** %s: %spassword %s\n",
				     _("Usage"), command_prefix,
				     _("<new password>"));
		return;
	}

	/* Update the password using prefs */
	send_enaml_buildf(ci->server->sock, "prefs", "password:%s", text);
}

/* user_status
 * Handle the /status command
 */
static void
user_status(ChatInterface *ci, const gchar *text)
{
	g_return_if_fail(ci);

	/* Send a status request */
	send_enaml_buildf(ci->server->sock, "status", "");
}

/* user_away
 * Handle the /away command
 */
static void
user_away(ChatInterface *ci, const gchar *text)
{
	g_return_if_fail(ci);

	/* Can do one of two things.  Set our away state and message or
	 * toggle our away state between away and active
	 */
	if (strlen(text)) {
		gchar *text2;

		text2 = utf8_encode(text);
		send_enaml_buildf(ci->server->sock, "status", "mode:%s text:%s",
				  "away", text2);
		g_free(text2);

	} else if (ci->server->me->status != SAY2_STATUS_AWAY) {
		send_enaml_buildf(ci->server->sock, "status", "mode:%s text:%s",
				  "away", "");
	} else {
		send_enaml_buildf(ci->server->sock, "status", "mode:%s text:%s",
				  "active", "");
	}
}

/* user_busy
 * Handle the /busy command
 */
static void
user_busy(ChatInterface *ci, const gchar *text)
{
	g_return_if_fail(ci);

	/* Can do one of two things.  Set our busy state and message or
	 * toggle our busy state between busy and active
	 */
	if (strlen(text)) {
		gchar *text2;

		text2 = utf8_encode(text);
		send_enaml_buildf(ci->server->sock, "status", "mode:%s text:%s",
				  "busy", text2);
		g_free(text2);

	} else if (ci->server->me->status != SAY2_STATUS_BUSY) {
		send_enaml_buildf(ci->server->sock, "status", "mode:%s text:%s",
				  "busy", "");
	} else {
		send_enaml_buildf(ci->server->sock, "status", "mode:%s text:%s",
				  "active", "");
	}
}

/* user_here
 * Handle the /here command
 */
static void
user_here(ChatInterface *ci, const gchar *text)
{
	g_return_if_fail(ci);

	/* Set our stats to here, with optional text message */
	if (strlen(text)) {
		gchar *text2;

		text2 = utf8_encode(text);
		send_enaml_buildf(ci->server->sock, "status", "mode:%s text:%s",
				  "active", text2);
		g_free(text2);
	} else {
		send_enaml_buildf(ci->server->sock, "status", "mode:%s text:%s",
				  "active", "");
	}
}

/* user_nick
 * Handle the /nick command
 */
static void
user_nick(ChatInterface *ci, const gchar *text)
{
	g_return_if_fail(ci);

	/* New nickname required */
	if (!strlen(text)) {
		chat_interface_sayto(ci, "*** %s: %snick %s\n",
				     _("Usage"), command_prefix,
				     _("<new nickname>"));
		return;
	}

	/* Send a change-nick request */
	send_enaml_buildf(ci->server->sock, "change-nick", "nick:%s", text);
}

/* user_whom
 * Handle the /whom command
 */
static void
user_whom(ChatInterface *ci, const gchar *text)
{
	g_return_if_fail(ci);

	/* Any argument is treated as a grep.  Whom gives a verbose
	 * almost-whois of each client returned and has to be handled
	 * specially.
	 */
	if (strlen(text)) {
		send_enaml_buildf(ci->server->sock, "list-clients",
				  "type:%s verbose grep:%s", "user", text);
	} else {
		send_enaml_buildf(ci->server->sock, "list-clients",
				  "type:%s verbose", "user");
	}
}

/* user_who
 * Handle the /who command
 */
static void
user_who(ChatInterface *ci, const gchar *text)
{
	g_return_if_fail(ci);

	/* Any argument is treated as a grep.  Who just returns a Members
	 * structure
	 */
	if (strlen(text)) {
		send_enaml_buildf(ci->server->sock, "list-clients",
				  "type:%s grep:%s", "user", text);
	} else {
		send_enaml_buildf(ci->server->sock, "list-clients",
				  "type:%s", "user");
	}
}

/* user_whois
 * Handle the /whois command
 */
static void
user_whois(ChatInterface *ci, const gchar *text)
{
	g_return_if_fail(ci);

	/* Need a nickname */
	if (!strlen(text)) {
		chat_interface_sayto(ci, "*** %s: %swhois %s\n",
				     _("Usage"), command_prefix,
				     _("<nickname>"));
		return;
	}

	/* Send a whois request for the online user given */
	send_enaml_buildf(ci->server->sock, "whois", "nick:%s type:%s",
			  text, "user");
}

/* user_seen
 * Handle the /seen command
 */
static void
user_seen(ChatInterface *ci, const gchar *text)
{
	g_return_if_fail(ci);

	/* Need an ID */
	if (!strlen(text)) {
		chat_interface_sayto(ci, "*** %s: %sseen %s\n",
				     _("Usage"), command_prefix,
				     _("<user id>"));
		return;
	}

	/* Send a whois request for the user id given */
	send_enaml_buildf(ci->server->sock, "whois", "id:%s type:%s",
			  text, "user");
}

/* user_channel
 * Handle the /channel command
 */
static void
user_channel(ChatInterface *ci, const gchar *text)
{
	g_return_if_fail(ci);

	/* Need an ID */
	if (!strlen(text)) {
		chat_interface_sayto(ci, "*** %s: %schannel %s\n",
				     _("Usage"), command_prefix,
				     _("<channel id>"));
		return;
	}

	/* Send a whois request for the channel id given */
	send_enaml_buildf(ci->server->sock, "whois", "id:%s type:%s",
			  text, "channel");
}

/* user_say
 * Handle the /say command
 */
static void
user_say(ChatInterface *ci, const gchar *text)
{
	gchar *text2;

	g_return_if_fail(ci);

	/* Need text to say */
	if (!strlen(text)) {
		chat_interface_sayto(ci, "*** %s: %ssay %s\n",
				     _("Usage"), command_prefix, _("<text>"));
		return;
	}

	/* If we were away, automatically bring us back.
	 * FIXME preference!
	 */
	if (ci->server->me->status == SAY2_STATUS_AWAY) {
		send_enaml_buildf(ci->server->sock, "status", "mode:%s text:%s",
				  "active", "");
	}

	/* Send the text to the server */
	text2 = utf8_encode(text);
	send_enaml_buildf(ci->server->sock, "say", "text:%s", text2);
	g_free(text2);
}

/* user_me
 * Handle the /me command
 */
static void
user_me(ChatInterface *ci, const gchar *text)
{
	gchar *text2;
	
	g_return_if_fail(ci);

	/* If the Recipient is a USER, /me actually does a /act */
	if (ci->recip_type == CHAT_RECIPTYPE_USER) {
		gchar *msg;

		/* Kludgy, but it works */
		msg = g_strdup_printf("%s %s", ci->recip->nick, text);
		user_act(ci, msg);
		g_free(msg);

		return;
	}

	/* Need text to say */
	if (!strlen(text)) {
		chat_interface_sayto(ci, "*** %s: %sme %s\n",
				     _("Usage"), command_prefix, _("<text>"));
		return;
	}

	/* If we were away, automatically bring us back.
	 * FIXME preference!
	 */
	if (ci->server->me->status == SAY2_STATUS_AWAY) {
		send_enaml_buildf(ci->server->sock, "status", "mode:%s text:%s",
				  "active", "");
	}

	/* Send the text to the server */
	text2 = utf8_encode(text);
	send_enaml_buildf(ci->server->sock, "say", "act:%s", text2);
	g_free(text2);
}

/* user_msg
 * Handle the /msg command
 */
static void
user_msg(ChatInterface *ci, const gchar *text)
{
	gchar **arg;
	gchar  *text2;

	g_return_if_fail(ci);

	/* Need two arguments, a nickname and some text to say */
	arg = g_strsplit(text, " ", 1);
	if (!arg[0] || !strlen(arg[0]) || !arg[1] || !strlen(arg[1])) {
		g_strfreev(arg);
		chat_interface_sayto(ci, "*** %s: %smsg %s\n",
				     _("Usage"), command_prefix,
				     _("<nickname> <text>"));
		return;
	}

	/* If we were away, automatically bring us back.
	 * FIXME preference!
	 */
	if (ci->server->me->status == SAY2_STATUS_AWAY) {
		send_enaml_buildf(ci->server->sock, "status", "mode:%s text:%s",
				  "active", "");
	}

	/* Send the text to the server */
	text2 = utf8_encode(arg[1]);
	send_enaml_buildf(ci->server->sock, "message", "nick:%s text:%s",
			  arg[0], text2);
	g_free(text2);

	g_strfreev(arg);
}

/* user_act
 * Handle the /act command
 */
static void
user_act(ChatInterface *ci, const gchar *text)
{
	gchar **arg;
	gchar  *text2;

	g_return_if_fail(ci);

	/* Need two arguments, a nickname and some text to say */
	arg = g_strsplit(text, " ", 1);
	if (!arg[0] || !strlen(arg[0]) || !arg[1] || !strlen(arg[1])) {
		g_strfreev(arg);
		chat_interface_sayto(ci, "*** %s: %sact %s\n",
				     _("Usage"), command_prefix,
				     _("<nickname> <text>"));
		return;
	}

	/* If we were away, automatically bring us back.
	 * FIXME preference!
	 */
	if (ci->server->me->status == SAY2_STATUS_AWAY) {
		send_enaml_buildf(ci->server->sock, "status", "mode:%s text:%s",
				  "active", "");
	}

	/* Send the text to the server */
	text2 = utf8_encode(arg[1]);
	send_enaml_buildf(ci->server->sock, "message", "nick:%s act:%s",
			  arg[0], text2);
	g_free(text2);

	g_strfreev(arg);
}

/* user_wall
 * Handle the /wall command
 */
static void
user_wall(ChatInterface *ci, const gchar *text)
{
	gchar *text2;

	g_return_if_fail(ci);

	/* Need some text to say */
	if (!strlen(text)) {
		chat_interface_sayto(ci, "*** %s: %swall %s\n",
				     _("Usage"), command_prefix, _("<text>"));
		return;
	}

	/* If we were away, automatically bring us back.
	 * FIXME preference!
	 */
	if (ci->server->me->status == SAY2_STATUS_AWAY) {
		send_enaml_buildf(ci->server->sock, "status", "mode:%s text:%s",
				  "active", "");
	}

	/* Send the text to the server */
	text2 = utf8_encode(text);
	send_enaml_buildf(ci->server->sock, "wall", "text:%s", text2);
	g_free(text2);
}

/* user_history
 * Handle the /history command
 */
static void
user_history(ChatInterface *ci, const gchar *text)
{
	g_return_if_fail(ci);

	/* Any argument is interpreted as a max indicator */
	if (strlen(text)) {
		send_enaml_buildf(ci->server->sock, "history", "max:%s", text);
	} else {
		send_enaml_buildf(ci->server->sock, "history", "");
	}
}

/* user_boot
 * Handle the /boot command
 */
static void
user_boot(ChatInterface *ci, const gchar *text)
{
	gchar **arg;

	g_return_if_fail(ci);

	/* First argument must be a nickname, the second, as reason, is
	 * optional
	 */
	arg = g_strsplit(text, " ", 1);
	if (!arg[0] || !strlen(arg[0])) {
		g_strfreev(arg);
		chat_interface_sayto(ci, "*** %s: %sboot %s\n",
				     _("Usage"), command_prefix,
				     _("<nickname> [ <reason> ]"));
		return;
	}

	/* Send the command to the server with the optional reason */
	if (arg[1] && strlen(arg[1])) {
		gchar *text2;

		text2 = utf8_encode(arg[1]);
		send_enaml_buildf(ci->server->sock, "boot-user",
				  "nick:%s text:%s", arg[0], text2);
		g_free(text2);
	} else {
		send_enaml_buildf(ci->server->sock, "boot-user", "nick:%s",
				  arg[0]);
	}
	
	g_strfreev(arg);
}

/* user_clear
 * Handle the /clear command
 */
static void
user_clear(ChatInterface *ci, const gchar *text)
{
	g_return_if_fail(ci);

	/* Clear the chat interfaces output box */
	gtk_editable_delete_text(GTK_EDITABLE(ci->output_box), 0, -1);
}

/* user_connect
 * Handle the /connect command
 */
static void
user_connect(ChatInterface *ci, const gchar *text)
{
	Say2Server  *server;
	gchar       *server_port;
	gchar      **arg;

	g_return_if_fail(ci);

	/* A /connect on its own means we'll pop open the ConnectDialog and
	 * ask for the relevant details
	 */
	if (!strlen(text)) {
		GtkWidget *cd, *win;

		/* Disconnect the interface from whatever it's connected to */
		chat_interface_disconnect(ci);

		/* Get the parent window of the ChatInterface */
		win = gtk_widget_get_ancestor(ci->top_vbox, GTK_TYPE_WINDOW);

		/* Open a ConnectDialog */
		cd = connect_dialog_new(win);
		gtk_widget_show(cd);
		return;
	}

	/* First argument must be the server_name, the second, the port, is
	 * optional.  The first can also be in the form server:port in which
	 * case we'll split that out
	 */
	arg = g_strsplit(text, " ", 1);
	if (!arg[0] || !strlen(arg[0])) {
		g_strfreev(arg);
		chat_interface_sayto(ci, "*** %s: %sconnect %s\n",
				     _("Usage"), command_prefix,
				     _("<hostname> [ <port> ]"));
		return;
	}

	/* If the port is ommited, the server_name may contain a : followed
	 * by the port
	 */
	if (!arg[1] || !strlen(arg[1])) {
		if ((server_port = strchr(arg[0], ':'))) {
			*server_port = 0;
			server_port++;

			if (!strlen(server_port))
				server_port = 0;
		}
	} else {
		server_port = arg[1];
	}

	/* Make the server connection */
	server = say2_server_new(arg[0], server_port);
	if (!server)
		return;

	/* Point the ChatInterface at it */
	chat_interface_set(ci, server, CHAT_RECIPTYPE_ANY, NULL);

	g_strfreev(arg);
}

/* user_disconnect
 * Handle the /disconnect command
 */
static void
user_disconnect(ChatInterface *ci, const gchar *text)
{
	g_return_if_fail(ci);
	
	/* This could be called when we have no server */
	if (!ci->server) {
		chat_interface_sayto(ci, "*** %s\n", _("Not connected"));
		return;
	}

	/* Call close on all the chat interfaces with this server */
	chat_interface_update(ci->server, CHAT_UPDATE_CLOSE, NULL);
}

/* user_login
 * Handle the /login command
 */
static void
user_login(ChatInterface *ci, const gchar *text)
{
	gchar **arg;

	g_return_if_fail(ci);

	/* This could be called when we have no server, we can't login then. */
	if (!ci->server) {
		chat_interface_sayto(ci, "*** %s\n",
				     _("Not connected, use /connect"));
		return;
	}

	/* Two arguments needed, a nickname and a password */
	arg = g_strsplit(text, " ", 1);
	if (!arg[0] || !strlen(arg[0]) || !arg[1] || !strlen(arg[1])) {
		g_strfreev(arg);
		chat_interface_sayto(ci, "*** %s: %slogin %s\n",
				     _("Usage"), command_prefix,
				     _("<user id> <password>"));
		return;
	}

	/* Set the login */
	say2_server_set_login(ci->server, arg[0], arg[1]);

	g_strfreev(arg);
}

/* user_guest
 * Handle the /guest command
 */
static void
user_guest(ChatInterface *ci, const gchar *text)
{
	gchar **arg;

	g_return_if_fail(ci);

	/* This could be called when we have no server, we can't login then. */
	if (!ci->server) {
		chat_interface_sayto(ci, "*** %s\n",
				     _("Not connected, use /connect"));
		return;
	}

	/* Two arguments needed, a nickname and an e-mail address.  The third,
	 * a tagline, is optional
	 */
	arg = g_strsplit(text, " ", 2);
	if (!arg[0] || !strlen(arg[0]) || !arg[1] || !strlen(arg[1])) {
		g_strfreev(arg);
		chat_interface_sayto(ci, "*** %s: %sguest %s\n",
				     _("Usage"), command_prefix,
				     _("<nickname> <email> [ <tagline> ]"));
		return;
	}

	/* Set the login */
	say2_server_set_guest(ci->server, arg[0], arg[1], arg[2]);

	g_strfreev(arg);
}

/* user_close
 * Handle the /close command
 */
static void
user_close(ChatInterface *ci, const gchar *text)
{
	GtkWidget *win;
	
	g_return_if_fail(ci);

	/* Get the parent of the ChatInterface */
	win = gtk_widget_get_ancestor(ci->top_vbox, GTK_TYPE_WINDOW);
	if (win)
		gtk_widget_destroy(win);
}

/* user_quit
 * Handle the /quit command
 */
static void
user_quit(ChatInterface *ci, const gchar *text)
{
	g_return_if_fail(ci);

	/* Not much to this one */
	gtk_main_quit();
}


/* user_handle_text
 * Handle text entered by the user, either in the entry box or "automatically
 * entered" for them when they clicked a button or popup menu entry etc.
 */
void
user_handle_text(ChatInterface *ci, const gchar *text)
{
	static void *dispatch_table = NULL;
	static void *uc_dispatch_table = NULL;
	void        *table;

	g_return_if_fail(ci);

	/* Oh god, I'm turning into poskanser */
	if (!dispatch_table)
		dispatch_table = dispatch_create(command_table);
	if (!uc_dispatch_table)
		uc_dispatch_table = dispatch_create(uc_command_table);

	/* If the chat interface is connected, then the full range of commands
	 * is available.  Otherwise only a select few are.
	 */
	table = (ci->server && ci->server->state == SAY2_SERVER_ACTIVE
		 ? dispatch_table : uc_dispatch_table);

	/* If the text begins with the command prefix, treat it as a command
	 * and dispatch accordingly.  Otherwise attempt to manually dispatch
	 * and if all else fails just swallow it.
	 */
	if (!strncmp(text, command_prefix, strlen(command_prefix))) {
		gchar    *cmd;
		size_t    cmdlen;

		/* Copy the command out */
		text += strlen(command_prefix);
		cmdlen = strcspn(text, " ");
		cmd = g_strndup(text, cmdlen);

		/* Find the start of the text */
		text += cmdlen;
		if (*text)
			text++;
	
		/* Try to dispatch it, supplying the ChatInterface and rest
		 * of the text as the arguments
		 */
		if (!dispatch_go(table, cmd, ci, (void *)text, NULL)) {
			chat_interface_sayto(ci, "*** %s: %s\n",
					     _("Unknown command"), cmd);
		}

	} else if (strlen(text) && (ci->recip_type == CHAT_RECIPTYPE_CHAN)) {
		/* Not a command, the recipient of the ChatInterface is a
		 * channel, so it's a /say.  Dispatch manually.
		 */
		dispatch_go(table, "say", ci, (void *)text, NULL);

	} else if (strlen(text) && (ci->recip_type == CHAT_RECIPTYPE_USER)) {
		/* Not a command, the recipient of the ChatInterface is a
		 * person, so it's a /msg.  Dispatch manually.
		 */
		gchar *msg;

		/* Need to but the recipient's nickname on the front */
		msg = g_strdup_printf("%s %s", ci->recip->nick, text);
		dispatch_go(table, "msg", ci, (void *)msg, NULL);
		g_free(msg);

	}
}
